from qualang_tools.wirer.connectivity.channel_spec import ChannelSpec
from qualang_tools.wirer.instruments.instrument_channel import (
    InstrumentChannel,
    InstrumentChannelLfFemInput,
    InstrumentChannelLfFemOutput,
    InstrumentChannelMwFemInput,
    InstrumentChannelMwFemOutput,
    InstrumentChannelOpxPlusInput,
    InstrumentChannelOpxPlusOutput,
    InstrumentChannelOctaveInput,
    InstrumentChannelOctaveOutput,
    InstrumentChannelExternalMixerInput,
    InstrumentChannelExternalMixerOutput,
    InstrumentChannelOpxPlusDigitalOutput,
    InstrumentChannelMwFemDigitalOutput,
    InstrumentChannelLfFemDigitalOutput,
    InstrumentChannelOctaveDigitalInput,
    InstrumentChannelExternalMixerDigitalInput,
)

# A channel template is a partially filled InstrumentChannel object
ChannelTemplate = InstrumentChannel


class ChannelSpecMwFemSingle(ChannelSpec):
    def __init__(self, con: int = None, slot: int = None, in_port: int = None, out_port: int = None):
        """
        Represents a single-channel specification for Microwave Front-End Module (MW-FEM).

        If specified, any of the provided arguments can be used to constrain the configuration of a
        single MW-FEM channel. If no arguments are provided, the configuration is unconstrained.

        For example, if the argument slot=2 is specified, then this specification can ONLY be
        satisfied by an available MW-FEM installed in slot 2 of an OPX1000 chassis.

        Attributes:
            channel_templates (list): A list containing initialized
                InstrumentChannelMwFemInput and InstrumentChannelMwFemOutput objects.

        Args:
            con (int, optional): Controller identifier for both input and output channels (any positive integer).
            slot (int, optional): FEM slot identifier for both input and output channels (1–8).
            in_port (int, optional): Input port identifier (1–2).
            out_port (int, optional): Output port identifier (1–8).
        """

        super().__init__()
        self.channel_templates = [
            InstrumentChannelMwFemInput(con=con, slot=slot, port=in_port),
            InstrumentChannelMwFemOutput(con=con, slot=slot, port=out_port),
        ]


class ChannelSpecLfFemSingle(ChannelSpec):
    def __init__(
        self, con: int = None, in_slot: int = None, in_port: int = None, out_slot: int = None, out_port: int = None
    ):
        """
        Represents a single-channel specification for Low-Frequency Front-End Module (LF-FEM).

        If specified, any of the provided arguments can be used to constrain the configuration of a
        single LF-FEM channel. If no arguments are provided, the configuration is unconstrained.

        For example, if in_slot=3 is specified AND this specification is used later to allocate an input
        channel, then it should ONLY be satisfied by an LF-FEM (with a free input channel) installed in
        slot 3 of an OPX1000 chassis. If unspecified, it can be satisfied by an LF-FEM installed on any
        given slot.

        Attributes:
            channel_templates (list): A list containing initialized
                InstrumentChannelLfFemInput and InstrumentChannelLfFemOutput objects.

        Args:
            con (int, optional): Controller identifier common to both input and output channels (any positive integer).
            in_slot (int, optional): Slot index for the input channel (1–8; only relevant if allocating input).
            in_port (int, optional): Port index for the input channel (1–2; only relevant if allocating input).
            out_slot (int, optional): Slot index for the output channel (1–8; only relevant if allocating output).
            out_port (int, optional): Port index for the output channel (1–8; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=in_slot, port=in_port),
            InstrumentChannelLfFemOutput(con=con, slot=out_slot, port=out_port),
        ]


class ChannelSpecLfFemBaseband(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        slot: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
    ):
        """
        Represents a baseband (unmixed) I/Q channel specification for LF-FEM.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        baseband I/Q LF-FEM channel group. If no arguments are provided, the configuration is unconstrained.

        For example, if in_port_q=2 is specified AND this specification is later used to allocate an
        input Q-quadrature channel, then it should ONLY be satisfied by an LF-FEM with its Q input on port 2.
        If unspecified, it can be satisfied by any available port on the FEM.

        Attributes:
            channel_templates (list): Two LF-FEM input and two LF-FEM output channels for I and Q.

        Args:
            con (int, optional): Controller identifier (any positive integer).
            slot (int, optional): Slot index shared by all four I/Q channels (1–8).
            in_port_i (int, optional): Port index for the I input channel (1–2; only relevant if allocating input).
            in_port_q (int, optional): Port index for the Q input channel (1–2; only relevant if allocating input).
            out_port_i (int, optional): Port index for the I output channel (1–8; only relevant if allocating output).
            out_port_q (int, optional): Port index for the Q output channel (1–8; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_i),
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_q),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_i),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_q),
        ]


class ChannelSpecOpxPlusSingle(ChannelSpec):
    def __init__(self, con: int = None, in_port: int = None, out_port: int = None):
        """
        Represents a single-channel specification for OPX+.

        If specified, any of the provided arguments can be used to constrain the configuration of an
        OPX+ channel. If no arguments are provided, the configuration is unconstrained.

        For example, if out_port=1 is specified AND this specification is used to allocate an output
        channel, then it should ONLY be satisfied by an OPX+ with output port 1 available. If unspecified,
        then the specification can be satisfied by any available port.

        Attributes:
           channel_templates (list): One OPX+ input and one OPX+ output channel.

        Args:
           con (int, optional): Controller (OPX+) identifier (any positive integer).
           in_port (int, optional): Port index for input (1-2; only relevant if allocating input).
           out_port (int, optional): Port index for output (1-10; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port),
        ]


class ChannelSpecOpxPlusBaseband(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
    ):
        """
        Represents a baseband (unmixed) I/Q channel specification for OPX+.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        baseband I/Q OPX+ channel group. If no arguments are provided, the configuration is unconstrained.

        For example, if out_port_i=1 is specified AND this specification is later used to allocate an
        I-quadrature output channel, then it should ONLY be satisfied by an OPX+ with output port 1 available.

        Attributes:
            channel_templates (list): I/Q-paired OPX+ input and output channels.

        Args:
            con (int, optional): Controller identifier.
            in_port_i (int, optional): Input port index for I (1-2; only relevant if allocating input).
            in_port_q (int, optional): Input port index for Q (1-2; only relevant if allocating input).
            out_port_i (int, optional): Output port index for I (1-10; only relevant if allocating output).
            out_port_q (int, optional): Output port index for Q (1-10; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port_i),
            InstrumentChannelOpxPlusInput(con=con, port=in_port_q),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_i),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_q),
        ]


class ChannelSpecOctave(ChannelSpec):
    def __init__(self, index: int = None, rf_in: int = None, rf_out: int = None):
        """
        Represents an RF channel specification for an Octave module.

        If specified, any of the provided arguments can be used to constrain the configuration of an
        Octave RF channel. If no arguments are provided, the configuration is unconstrained.

        For example, if rf_out=1 is specified AND this specification is used to allocate an output channel,
        then it should ONLY be satisfied by an Octave with RF output port 1 available.

        Attributes:
            channel_templates (list): One Octave input and one Octave output channel.

        Args:
            index (int, optional): Octave module identifier (any positive integer).
            rf_in (int, optional): RF input port index (1-2; only relevant if allocating input).
            rf_out (int, optional): RF output port index (1-5; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOctaveInput(con=index, port=rf_in),
            InstrumentChannelOctaveOutput(con=index, port=rf_out),
        ]


class ChannelSpecExternalMixer(ChannelSpec):
    def __init__(self, index: int = None, rf_in: int = None, rf_out: int = None):
        """
        Represents an RF channel specification for an External Mixer.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        single RF channel. If no arguments are provided, the configuration is unconstrained.

        For example, if rf_out=0 is specified AND this specification is used to allocate an output
        channel, then it should ONLY be satisfied by an External Mixer with RF output port 0 available.

        Args:
            index (int, optional): External mixer identifier.
            rf_in (int, optional): Input RF port (only relevant if allocating input).
            rf_out (int, optional): Output RF port (only relevant if allocating output).
        """

        super().__init__()
        self.channel_templates = [
            InstrumentChannelExternalMixerInput(con=index, port=rf_in),
            InstrumentChannelExternalMixerOutput(con=index, port=rf_out),
        ]


class ChannelSpecLfFemBasebandAndOctave(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        slot: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        octave_index: int = None,
        rf_in: int = None,
        rf_out: int = None,
    ):
        """
        Represents a combined baseband I/Q + RF specification for LF-FEM and Octave.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        baseband I/Q and RF channel pair. If no arguments are provided, the configuration is unconstrained.

        For example, if rf_in=2 is specified AND this specification is used to allocate an input RF channel,
        then it should ONLY be satisfied by an Octave with input port 2 available. If it is unspecified,
        then the rf input channel can be satisfied by any available port.

        Note that if you only need input, then the `out` specifications will be ignored and vice versa.

        Args:
            con (int, optional): Controller identifier for the LF-FEM (any positive integer).
            slot (int, optional): Slot index shared by the I/Q baseband channels (1-8).
            in_port_i (int, optional): I input port index (1-2; only relevant if allocating input).
            in_port_q (int, optional): Q input port index (1-2; only relevant if allocating input).
            out_port_i (int, optional): I output port index (1-8; only relevant if allocating output).
            out_port_q (int, optional): Q output port index (1-8; only relevant if allocating output).
            octave_index (int, optional): Octave module index (any positive integer).
            rf_in (int, optional): RF input port (1-2; only relevant if allocating input).
            rf_out (int, optional): RF output port (1-5; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_i),
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_q),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_i),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_q),
            InstrumentChannelOctaveInput(con=octave_index, port=rf_in),
            InstrumentChannelOctaveOutput(con=octave_index, port=rf_out),
        ]


class ChannelSpecLfFemBasebandAndExternalMixer(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        slot: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        mixer_index: int = None,
    ):
        """
        Represents a combined baseband I/Q + RF specification for LF-FEM and External mixer.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        baseband I/Q and External mixer RF channel pair. If no arguments are provided, the configuration is unconstrained.

        For example, if mixer_index=1 is specified AND this specification is used to allocate an RF
        channel, then it should ONLY be satisfied by an external mixer with index 1. Otherwise, it can
        be satisfied by any available external mixer.

        Args:
            con (int, optional): Controller identifier for the LF-FEM (any positive integer).
            slot (int, optional): Slot index for the LF-FEM (1-8).
            in_port_i (int, optional): I input port index (1-2; only relevant if allocating input).
            in_port_q (int, optional): Q input port index (1-2; only relevant if allocating input).
            out_port_i (int, optional): I output port index (1-8; only relevant if allocating output).
            out_port_q (int, optional): Q output port index (1-8; only relevant if allocating output).
            mixer_index (int, optional): External mixer index.
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_i),
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_q),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_i),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_q),
            InstrumentChannelExternalMixerInput(con=mixer_index, port=1),
            InstrumentChannelExternalMixerOutput(con=mixer_index, port=1),
        ]


class ChannelSpecOpxPlusBasebandAndOctave(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        octave_index: int = None,
        rf_in: int = None,
        rf_out: int = None,
    ):
        """
        Represents a combined baseband I/Q + RF specification for OPX+ and Octave.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        baseband I/Q and RF channel pair. If no arguments are provided, the configuration is unconstrained.

        For example, if out_port_q=3 is specified AND this specification is used to allocate the Q output,
        then it should ONLY be satisfied by an OPX+ with Q output port 3.

        Args:
            con (int, optional): Controller identifier for the OPX+ (any positive integer).
            in_port_i (int, optional): I input port index (1-2; only relevant if allocating input).
            in_port_q (int, optional): Q input port index (1-2; only relevant if allocating input).
            out_port_i (int, optional): I output port index (1-10; only relevant if allocating output).
            out_port_q (int, optional): Q output port index (1-10; only relevant if allocating output).
            octave_index (int, optional): Octave module index (any positive integer).
            rf_in (int, optional): RF input port (1-2; only relevant if allocating input).
            rf_out (int, optional): RF output port (1-5; only relevant if allocating output).
        """
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port_i),
            InstrumentChannelOpxPlusInput(con=con, port=in_port_q),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_i),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_q),
            InstrumentChannelOctaveInput(con=octave_index, port=rf_in),
            InstrumentChannelOctaveOutput(con=octave_index, port=rf_out),
        ]


class ChannelSpecOpxPlusBasebandAndExternalMixer(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        mixer_index: int = None,
    ):
        """
        Represents a combined baseband I/Q + RF specification for OPX+ and External mixer.

        If specified, any of the provided arguments can be used to constrain the configuration of a
        baseband I/Q and External mixer RF channel pair. If no arguments are provided, the configuration is unconstrained.

        For example, if mixer_index=0 is specified, then it should ONLY be satisfied by an External mixer
        with index 0 during allocation.

        Args:
            con (int, optional): Controller identifier for the OPX+ (any positive integer).
            in_port_i (int, optional): I input port index (1-2; only relevant if allocating input).
            in_port_q (int, optional): Q input port index (1-2; only relevant if allocating input).
            out_port_i (int, optional): I output port index (1-10; only relevant if allocating output).
            out_port_q (int, optional): Q output port index (1-10; only relevant if allocating output).
            mixer_index (int, optional): External mixer index.
        """

        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port_i),
            InstrumentChannelOpxPlusInput(con=con, port=in_port_q),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_i),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_q),
            InstrumentChannelExternalMixerInput(con=mixer_index, port=1),
            InstrumentChannelExternalMixerOutput(con=mixer_index, port=1),
        ]


class ChannelSpecOpxPlusDigital(ChannelSpec):
    def __init__(self, con: int = None, out_port: int = None):
        """
        Represents a digital output specification for OPX+.

        If specified, the arguments constrain the configuration of a digital output channel.
        If no arguments are provided, the configuration is unconstrained.

        For example, if out_port=6 is specified, then it should ONLY be satisfied by an OPX+
        digital output port 6. Otherwise, any available OPX+ digital output port can be selected.

        Args:
            con (int, optional): Controller identifier for the OPX+ (any positive integer).
            out_port (int, optional): Digital output port index (1-10).
        """
        super().__init__()
        self.channel_templates = [InstrumentChannelOpxPlusDigitalOutput(con=con, port=out_port)]


class ChannelSpecMwFemDigital(ChannelSpec):
    def __init__(self, con: int = None, slot: int = None, out_port: int = None):
        """
        Represents a digital output specification for MW-FEM.

        If specified, the arguments constrain the configuration of a digital output channel.
        If no arguments are provided, the configuration is unconstrained.

        For example, if con=2 is specified, then it should ONLY be satisfied by the MW-FEM controller
        with index 2. Otherwise, any available MW-FEM controller can be used.

        Args:
            con (int, optional): Controller identifier for the OPX1000
            slot (int, optional): FEM slot identifier for the MW-FEM (1–8).
            out_port (int, optional): Digital output port index (1-8).
        """
        super().__init__()
        self.channel_templates = [InstrumentChannelMwFemDigitalOutput(con=con, slot=slot, port=out_port)]


class ChannelSpecLfFemDigital(ChannelSpec):
    def __init__(self, con: int = None, slot: int = None, out_port: int = None):
        """
        Represents a digital output specification for LF-FEM.

        If specified, the arguments constrain the configuration of a digital output channel.
        If no arguments are provided, the configuration is unconstrained.

        For example, if out_port=0 is specified, then it should ONLY be satisfied by the LF-FEM
        digital output port 0. Otherwise, any available port may be selected.

        Args:
            con (int, optional): Controller identifier for the OPX1000.
            slot (int, optional): FEM slot inside the chassis for the LF-FEM (1–8).
            out_port (int, optional): Digital output port index.
        """
        super().__init__()
        self.channel_templates = [InstrumentChannelLfFemDigitalOutput(con=con, slot=slot, port=out_port)]


class ChannelSpecOctaveDigital(ChannelSpec):
    def __init__(self, con: int = None, in_port: int = None):
        """
        Represents a digital input specification for an Octave module.

        If specified, the arguments constrain the configuration of a digital input channel.
        If no arguments are provided, the configuration is unconstrained.

        For example, if in_port=4 is specified, then it should ONLY be satisfied by an Octave
        digital input port 4. Otherwise, any available Octave input port may be used.

        Args:
            con (int, optional): Controller identifier for the Octave module.
            in_port (int, optional): Digital input port index.
        """
        super().__init__()
        self.channel_templates = [InstrumentChannelOctaveDigitalInput(con=con, port=in_port)]


class ChannelSpecExternalMixerDigital(ChannelSpec):
    def __init__(self, con: int = None, in_port: int = None):
        """
        Represents a digital input specification for an External mixer.

        If specified, the arguments constrain the configuration of a digital input channel.
        If no arguments are provided, the configuration is unconstrained.

        For example, if con=1 is specified, then it should ONLY be satisfied by the External mixer
        controller with index 1. Otherwise, any available mixer can satisfy the specification.

        Args:
            con (int, optional): Controller identifier for the External mixer.
            in_port (int, optional): Digital input port index.
        """
        super().__init__()
        self.channel_templates = [InstrumentChannelExternalMixerDigitalInput(con=con, port=in_port)]


mw_fem_spec = ChannelSpecMwFemSingle
lf_fem_spec = ChannelSpecLfFemSingle
lf_fem_iq_spec = ChannelSpecLfFemBaseband
lf_fem_iq_octave_spec = ChannelSpecLfFemBasebandAndOctave
lf_fem_iq_ext_mixer_spec = ChannelSpecLfFemBasebandAndExternalMixer
opx_spec = ChannelSpecOpxPlusSingle
opx_iq_spec = ChannelSpecOpxPlusBaseband
opx_iq_octave_spec = ChannelSpecOpxPlusBasebandAndOctave
opx_iq_ext_mixer_spec = ChannelSpecOpxPlusBasebandAndExternalMixer
octave_spec = ChannelSpecOctave
ext_mixer_spec = ChannelSpecExternalMixer
opx_dig_spec = ChannelSpecOpxPlusDigital
lf_fem_dig_spec = ChannelSpecLfFemDigital
mw_fem_dig_spec = ChannelSpecMwFemDigital



"""
The purpose of this module is to compile a sequence of wiring specifications
into potential channel specifications, then to allocate them to the first valid
combination of instrument channels.
"""

import copy
from typing import List

from .channel_specs import (
    ChannelSpecLfFemSingle,
    ChannelSpecOpxPlusSingle,
    ChannelSpecMwFemSingle,
    ChannelSpecLfFemBaseband,
    ChannelSpecOctave,
    ChannelSpecExternalMixer,
    ChannelSpecOpxPlusBaseband,
    ChannelSpecMwFemDigital,
    ChannelSpecLfFemDigital,
    ChannelSpecOctaveDigital,
    ChannelSpecOpxPlusDigital,
    ChannelSpecExternalMixerDigital,
)
from .wirer_assign_channels_to_spec import assign_channels_to_spec
from .wirer_exceptions import ConstraintsTooStrictException, NotEnoughChannelsException
from ..connectivity.channel_spec import ChannelSpec
from ..instruments import Instruments
from ..connectivity import Connectivity
from ..connectivity.wiring_spec import WiringSpec, WiringFrequency, WiringLineType


def allocate_wiring(
    connectivity: Connectivity,
    instruments: Instruments,
    block_used_channels: bool = True,
    clear_wiring_specifications: bool = True,
    observe_pulser_allocation: bool = False,
):
    """
    Allocates available instrument channels to quantum elements based on their wiring specifications.

    This function takes the wiring specifications from the provided `Connectivity` instance and attempts
    to allocate channels for each specification to the available instrument channels within the `Instruments` instance.

    The allocation is done in a sequence based on the line types specified in the `line_type_fill_order`. The function
    handles different frequency requirements, either DC or RF, and manages the reuse of available channels depending on
    the `block_used_channels` flag.

    The allocation process involves the following steps:
    1. The function iterates over the wiring specifications (`specs`) based on their `line_type`, attempting to allocate
       channels for each `WiringSpec`.
    2. For each specification, the relevant allocation function (`_allocate_wiring`) is called to allocate DC or RF channels.
    3. If `clear_wiring_specifications` is `True`, the specifications are cleared after the allocation.
    4. If `block_used_channels` is `False`, any channels that were previously used but are no longer allocated are returned
       to the available channels list.

    Args:
        connectivity (Connectivity): An instance of the `Connectivity` class containing the wiring specifications to be allocated.
        instruments (Instruments): An instance of the `Instruments` class that manages available and used channels.
        block_used_channels (bool, optional): If `True`, prevents previously used channels from being returned to the available pool.
            Defaults to `True`.
        clear_wiring_specifications (bool, optional): If `True`, clears the list of wiring specifications in `connectivity`
            after allocation. Defaults to `True`.
        observe_pulser_allocation (bool, optional): If `True`, enforces pulser allocation constraints (16 pulsers per FEM,
            18 per OPX+). If `False`, ignores pulser limits and only respects physical channel limits.
            Defaults to `False`.

    Raises:
        ConstraintsTooStrictException: If the constraints for a wiring specification are too strict, preventing allocation.
        NotEnoughChannelsException: If there are not enough available channels to satisfy the wiring specification.
    """

    line_type_fill_order = [
        WiringLineType.RESONATOR,
        WiringLineType.DRIVE,
        WiringLineType.FLUX,
        WiringLineType.CHARGE,
        WiringLineType.COUPLER,
        WiringLineType.CROSS_RESONANCE,
        WiringLineType.ZZ_DRIVE,
    ]

    specs = connectivity.specs

    used_channel_cache = copy.deepcopy(instruments.used_channels)

    specs_with_untyped_lines = []
    for line_type in line_type_fill_order:
        for spec in specs:
            if spec.line_type not in line_type_fill_order:
                specs_with_untyped_lines.append(spec)
            if spec.line_type == line_type:
                _allocate_wiring(spec, instruments, observe_pulser_allocation)

    # remove duplicates
    specs_with_untyped_lines = list(dict.fromkeys(specs_with_untyped_lines))
    for spec in specs_with_untyped_lines:
        _allocate_wiring(spec, instruments, observe_pulser_allocation)

    if clear_wiring_specifications:
        connectivity.specs = []

    if not block_used_channels:
        for channel_type, used_channels in instruments.used_channels.items():
            for i, used_channel in enumerate(reversed(used_channels)):
                if used_channel not in used_channel_cache:
                    instruments.available_channels.insert(0, used_channel)
                    instruments.used_channels.remove(used_channel)


def _allocate_wiring(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    if spec.frequency == WiringFrequency.DC:
        allocate_dc_channels(spec, instruments, observe_pulser_allocation)

    elif spec.frequency == WiringFrequency.RF:
        allocate_rf_channels(spec, instruments, observe_pulser_allocation)

    elif spec.frequency == WiringFrequency.DO:
        allocate_do_channels(spec, instruments, observe_pulser_allocation)

    else:
        raise NotImplementedError()


def allocate_dc_channels(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    """
    Try to allocate DC channels to an LF-FEM or OPX+ to satisfy the spec.
    """
    dc_specs = [
        # LF-FEM, Single analog output
        ChannelSpecLfFemSingle() & ChannelSpecLfFemDigital(),
        # OPX+, Single analog output
        ChannelSpecOpxPlusSingle() & ChannelSpecOpxPlusDigital(),
    ]

    allocate_channels(
        spec, dc_specs, instruments, same_con=True, same_slot=True, observe_pulser_allocation=observe_pulser_allocation
    )


def allocate_rf_channels(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    """
    Try to allocate RF channels to a MW-FEM. If that doesn't work, look for a
    combination of LF-FEM I/Q and Octave channels, or OPX+ I/Q and Octave
    channels.
    """
    rf_specs = [
        # MW-FEM, Single RF output
        ChannelSpecMwFemSingle() & ChannelSpecMwFemDigital(),
        # LF-FEM I/Q output with Octave for upconversion
        ChannelSpecLfFemBaseband() & ChannelSpecLfFemDigital() & ChannelSpecOctave() & ChannelSpecOctaveDigital(),
        # LF-FEM I/Q output with External Mixer for upconversion
        ChannelSpecLfFemBaseband()
        & ChannelSpecLfFemDigital()
        & ChannelSpecExternalMixer()
        & ChannelSpecExternalMixerDigital(),
        # OPX+ I/Q output with Octave for upconversion
        ChannelSpecOpxPlusBaseband() & ChannelSpecOpxPlusDigital() & ChannelSpecOctave() & ChannelSpecOctaveDigital(),
        # OPX+ I/Q output with External Mixer for upconversion
        ChannelSpecOpxPlusBaseband()
        & ChannelSpecOpxPlusDigital()
        & ChannelSpecExternalMixer()
        & ChannelSpecExternalMixerDigital(),
    ]

    allocate_channels(
        spec, rf_specs, instruments, same_con=True, same_slot=True, observe_pulser_allocation=observe_pulser_allocation
    )


def allocate_do_channels(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    """
    Try to allocate Digital Only (DO) channels to an LF- or MW-FEM or OPX+ to satisfy the spec.
    """
    do_specs = [
        # LF-FEM, Single digital output
        ChannelSpecLfFemDigital(),
        # MW-FEM, Single digital output
        ChannelSpecMwFemDigital(),
        # OPX+, Single digital output
        ChannelSpecOpxPlusDigital(),
    ]

    allocate_channels(
        spec, do_specs, instruments, same_con=True, same_slot=True, observe_pulser_allocation=observe_pulser_allocation
    )


def allocate_channels(
    wiring_spec: WiringSpec,
    channel_specs: List[ChannelSpec],
    instruments: Instruments,
    same_con: bool,
    same_slot: bool,
    observe_pulser_allocation: bool = False,
):
    mask_failures = 0
    for channel_spec in channel_specs:
        channel_spec = channel_spec.filter_by_wiring_spec(wiring_spec)
        if wiring_spec.constraints:
            constraints = wiring_spec.constraints.filter_by_wiring_spec(wiring_spec)
            if not channel_spec.apply_constraints(constraints):
                mask_failures += 1
                continue
        if assign_channels_to_spec(
            wiring_spec,
            instruments,
            channel_spec.channel_templates,
            same_con=same_con,
            same_slot=same_slot,
            observe_pulser_allocation=observe_pulser_allocation,
        ):
            return

    if mask_failures == len(channel_specs):
        raise ConstraintsTooStrictException(wiring_spec, wiring_spec.constraints)
    else:
        raise NotEnoughChannelsException(wiring_spec)



import threading
from datetime import datetime
from time import sleep
from tqdm import tqdm


def run_in_thread(fn):
    def run(*k, **kw):
        t = threading.Thread(target=fn, args=k, kwargs=kw)
        t.start()

    return run


def pbar(res_handles, n_avg, n_label, timeout=10, return_times=False):
    n_now = 0
    n = None
    m = 0
    while n is None:
        if m * 0.1 > timeout:
            print("reached timeout")
            break
        sleep(0.1)
        n = res_handles.get(n_label).fetch_all()

    times_vec = []
    with tqdm(total=n_avg, desc=n_label) as pbar_obj:
        while n < n_avg:
            n = res_handles.get(n_label).fetch_all() + 1
            sleep(0.1)
            if n is not None and n > n_now:
                pbar_obj.update(n - n_now)
                n_now = n
                if return_times:
                    times_vec.append(datetime.now())
    if return_times:
        return times_vec



import numpy as np
from qm.qua import *

from qualang_tools.bakery.bakery import baking
from qualang_tools.bakery.randomized_benchmark_c1 import c1_ops, c1_table


class RBOneQubit:
    def __init__(self, config: dict, d_max: int, K: int, qubit: str):
        """
        Class to retrieve easily baked RB sequences and their inverse operations

        :param config: Configuration file
        :param d_max: Maximum length of desired RB sequence
        :param K: Number of RB sequences
        :param qubit: Name of the quantum element designating the qubit
        """
        if not (qubit in config["elements"]):
            raise KeyError(f"Quantum element {qubit} is not in the config")

        self.sequences = [RBSequence(config, d_max, qubit) for _ in range(K)]
        self.inverse_ops = [seq.revert_ops for seq in self.sequences]
        self.duration_trackers = [seq.duration_tracker for seq in self.sequences]
        self.baked_sequences = [seq.sequence for seq in self.sequences]


def find_revert_op(input_state_index: int):
    """Looks in the Cayley table the operation needed to reset the state to ground state from input state_tracker

    :param input_state_index: Index of the current state tracker
    :return: index of the next Clifford to apply to invert RB sequence
    """
    for i in range(len(c1_ops)):
        if c1_table[input_state_index][i] == 0:
            return i


def play_revert_op(index: int, baked_cliffords):
    """
    Plays an operation resetting qubit in its ground state based on the
    transformation provided by the index in Cayley table (switch using baked Cliffords)

    :param baked_cliffords: list of baked cliffords (generated using method generate_cliffords)
    :param index: index of the transformed qubit state
    """

    with switch_(index):
        for i in range(len(baked_cliffords)):
            with case_(i):
                baked_cliffords[i].run()


class RBSequence:
    def __init__(self, config: dict, d_max: int, qubit: str):
        self.d_max = d_max
        self.config = config
        self.qubit = qubit
        self.state_tracker = [0] * d_max  # Keeps track of all transformations done on qubit state
        self.state_init = 0
        self.revert_ops = [0] * d_max  # Keeps track of inverse op index associated to each sequence
        self.duration_tracker = [0] * d_max  # Keeps track of each Clifford's duration
        # self.baked_cliffords = self.generate_cliffords()  # List of baking objects for running Cliffords
        self.operations_list = [None] * d_max
        self.inverse_op_string = [""] * d_max
        self.sequence = self.generate_RB_sequence()  # Store the RB sequence

    def play_revert_op2(self, index: int):
        """
        Plays an operation resetting qubit in its ground state based on the
        transformation provided by the index in Cayley table (explicit switch case)

        :param index: index of the transformed qubit state
        """
        qubit = self.qubit

        with switch_(index):
            with case_(0):
                play("I", qubit)
            with case_(1):
                play("X", qubit)
            with case_(2):
                play("Y", qubit)
            with case_(3):
                play("Y", qubit)
                play("X", qubit)
            with case_(4):
                play("X/2", qubit)
                play("Y/2", qubit)
            with case_(5):
                play("X/2", qubit)
                play("-Y/2", qubit)
            with case_(6):
                play("-X/2", qubit)
                play("Y/2", qubit)
            with case_(7):
                play("-X/2", qubit)
                play("-Y/2", qubit)
            with case_(8):
                play("Y/2", qubit)
                play("X/2", qubit)
            with case_(9):
                play("Y/2", qubit)
                play("-X/2", qubit)
            with case_(10):
                play("-Y/2", qubit)
                play("X/2", qubit)
            with case_(11):
                play("-Y/2", qubit)
                play("-X/2", qubit)
            with case_(12):
                play("X/2", qubit)
            with case_(13):
                play("-X/2", qubit)
            with case_(14):
                play("Y/2", qubit)
            with case_(15):
                play("-Y/2", qubit)
            with case_(16):
                play("-X/2", qubit)
                play("Y/2", qubit)
                play("X/2", qubit)
            with case_(17):
                play("-X/2", qubit)
                play("-Y/2", qubit)
                play("X/2", qubit)
            with case_(18):
                play("X", qubit)
                play("Y/2", qubit)
            with case_(19):
                play("X", qubit)
                play("-Y/2", qubit)
            with case_(20):
                play("Y", qubit)
                play("X/2", qubit)
            with case_(21):
                play("Y", qubit)
                play("-X/2", qubit)
            with case_(22):
                play("X/2", qubit)
                play("Y/2", qubit)
                play("X/2", qubit)
            with case_(23):
                play("-X/2", qubit)
                play("Y/2", qubit)
                play("-X/2", qubit)

    def generate_cliffords(self):
        """
        Returns a list of baking object giving access to baked Clifford waveforms

        :return: List of baking objects to play each Clifford
        """

        baked_clifford = [None] * len(c1_ops)
        for i in range(len(c1_ops)):
            with baking(self.config) as b2:
                for op in c1_ops[i]:
                    b2.play(op, self.qubit)
            baked_clifford[i] = b2
        return baked_clifford

    def generate_RB_sequence(self):
        """
        Creates a baking object generating a random Clifford sequence of length d_max
        """
        with baking(self.config) as b:
            for d in range(self.d_max):
                i = np.random.randint(0, len(c1_ops))
                if d > 0:
                    self.duration_tracker[d] = self.duration_tracker[
                        d - 1
                    ]  # Set duration to value of the sequence step

                # Play the random Clifford
                random_clifford = c1_ops[i]
                self.operations_list[d] = random_clifford
                for op in random_clifford:
                    b.play(op, self.qubit)
                    self.duration_tracker[d] += 1  # Add additional duration for each pulse played to build Clifford

                if d == 0:  # Handle the case for qubit set to original/ground state
                    self.state_tracker[d] = c1_table[self.state_init][i]
                else:  # Get the newly transformed state within th Cayley table based on previous step
                    self.state_tracker[d] = c1_table[self.state_tracker[d - 1]][i]
                self.revert_ops[d] = find_revert_op(self.state_tracker[d])
                self.inverse_op_string[d] = c1_ops[self.revert_ops[d]]
        return b



from dataclasses import dataclass
from typing import List, Dict, Union

from .wiring_spec import WiringLineType
from ..instruments.instrument_channel import AnyInstrumentChannel


@dataclass(frozen=True)
class Reference:
    name: str

    def __str__(self):
        return self.name


@dataclass(frozen=True)
class QubitReference:
    index: int

    def __str__(self):
        return f"q{self.index}"


@dataclass(frozen=True)
class QubitPairReference:
    control_index: int
    target_index: int

    def __str__(self):
        return f"q{self.control_index}-{self.target_index}"


ElementId = Union[Reference, QubitReference, QubitPairReference]


class Element:
    def __init__(self, id: Union[str, QubitReference, QubitPairReference]):
        if isinstance(id, str):
            id = Reference(id)
        self.id = id
        self.channels: Dict[WiringLineType, List[AnyInstrumentChannel]] = dict()

    def __str__(self):
        return str(self.channels)

    def __eq__(self, other):
        return self.id == other.id



import numpy as np

from qm.qua import declare, assign, play, fixed, Cast, amp, wait, ramp, ramp_to_zero, Math, if_, elif_
from typing import Union, List, Dict
from warnings import warn
from qm.qua._expressions import QuaExpression, QuaVariable


class VoltageGateSequence:
    def __init__(self, configuration: Dict, elements: List[str]):
        """
        Initializes a VirtualGateSequence object for designing arbitrary pulse sequences using virtual gates.

        This class allows the creation of complex pulse sequences by defining voltage levels (points) in gate-space
        and durations for multiple elements. It supports the addition of steps and ramps, and keeps track of average
        voltage levels for compensation pulses.

        The `configuration` provided will be updated to include necessary operations and waveforms for the sequence.

        **Warning: The framework and compensation pulse derivation is working only for sequences shorter than 8ms.**
        :param configuration: A dictionary representing the OPX configuration (this will be modified)
        :param elements: A list of elements (strings) involved in the virtual gate operations.
        """
        # List of the elements involved in the virtual gates
        self._elements = elements
        # The OPX configuration
        self._config = configuration
        # Initialize the current voltage level for sticky elements
        self.current_level = [0.0 for _ in self._elements]
        # Relevant voltage points in the charge stability diagram
        self._voltage_points = {}
        # Keep track of the averaged voltage played for defining the compensation pulse at the end of the sequence
        self.average_power = [0 for _ in self._elements]
        self._expression = None
        self._expression2 = None
        self._voltage_tolerance = 0.001  # limit for adding compensation. must be positive.
        self.base_operation = {}
        # Add to the config the step operation (length=16ns & amp=0.25V)
        for el in self._elements:
            base_amplitude = self._check_amplified_mode(el)
            self.base_operation[el] = {"amplitude": base_amplitude, "bit_shift": int(np.log2(1 / base_amplitude))}
            self._config["elements"][el]["operations"]["step"] = f"{el}_step_pulse"
            self._config["pulses"][f"{el}_step_pulse"] = {
                "operation": "control",
                "length": 16,
                "waveforms": {"single": f"{el}_step_wf"},
            }
            self._config["waveforms"][f"{el}_step_wf"] = {"type": "constant", "sample": base_amplitude}

    def _check_amplified_mode(self, element: str):
        con = self._config["elements"][element]["singleInput"]["port"][0]
        if "type" in self._config["controllers"][con].keys():
            if self._config["controllers"][con]["type"] == "opx1000":
                fem = self._config["elements"][element]["singleInput"]["port"][1]
                ch = self._config["elements"][element]["singleInput"]["port"][2]
                if self._config["controllers"][con]["fems"][fem]["type"] == "LF":
                    if (
                        self._config["controllers"][con]["fems"][fem]["analog_outputs"][ch]["output_mode"]
                        == "amplified"
                    ):
                        return 1.0

        return 0.25

    def _check_name(self, name, key):
        if name in key:
            return self._check_name(name + "%", key)
        else:
            return name

    def _add_op_to_config(self, el: str, name: str, amplitude: float, length: int) -> str:
        """Add an operation to an element when the amplitude is fixed to release the number of real-time operations on
        the OPX.

        :param el: the element to which we want to add the operation.
        :param name: name of the operation.
        :param amplitude: Amplitude of the pulse in V.
        :param length: Duration of the pulse in ns.
        :return : The name of the created operation.
        """
        op_name = self._check_name(name, self._config["elements"][el]["operations"])
        pulse_name = self._check_name(f"{el}_{op_name}_pulse", self._config["pulses"])
        wf_name = self._check_name(f"{el}_{op_name}_wf", self._config["waveforms"])
        self._config["elements"][el]["operations"][op_name] = pulse_name
        self._config["pulses"][pulse_name] = {
            "operation": "control",
            "length": length,
            "waveforms": {"single": wf_name},
        }
        self._config["waveforms"][wf_name] = {"type": "constant", "sample": amplitude}
        return op_name

    @staticmethod
    def _check_duration(duration: int):
        if duration is not None and not isinstance(duration, (QuaExpression, QuaVariable)):
            if duration == 0:
                warn(
                    "\nThe duration of one level is set to zero which can cause gaps, use with care or set it it to at least 16ns.",
                    stacklevel=2,
                )
            else:
                assert duration >= 4, "The duration must be a larger than 16 ns."
                assert duration % 4 == 0, "The duration must be a multiple integer of 4ns."

    def _update_averaged_power(self, level, duration, ramp_duration=None, current_level=None):
        if self.is_QUA(level):
            self._expression = declare(fixed)
            assign(self._expression, level)
            # Multiplication by 1024 to increase the resolution since duration*level must be an integer
            new_average = Cast.mul_int_by_fixed(duration << 10, self._expression)
        elif self.is_QUA(duration):
            new_average = Cast.mul_int_by_fixed(duration << 10, float(level))
        else:
            new_average = int(np.round((level * duration) * 1024))

        if ramp_duration is not None:
            if not self.is_QUA(ramp_duration):
                if self.is_QUA(level):
                    self._expression2 = declare(fixed)
                    assign(self._expression2, (self._expression + current_level) >> 1)
                    new_average += Cast.mul_int_by_fixed(ramp_duration << 10, self._expression2)
                elif self.is_QUA(current_level):
                    expression2 = declare(fixed)
                    assign(expression2, (level + current_level) >> 1)
                    new_average += Cast.mul_int_by_fixed(ramp_duration << 10, expression2)
                elif self.is_QUA(duration):
                    new_average += Cast.mul_int_by_fixed(ramp_duration << 10, (level + current_level) / 2)
                else:
                    new_average += int(np.round(1024 * (level + current_level) * ramp_duration / 2))

            else:
                new_average += Cast.mul_int_by_fixed(ramp_duration << 10, (level + current_level) / 2)
        return new_average

    @staticmethod
    def is_QUA(var):
        return isinstance(var, (QuaExpression, QuaVariable))

    def add_step(
        self,
        level: list[Union[float, QuaExpression, QuaVariable]] = None,
        duration: Union[int, QuaExpression, QuaVariable] = None,
        voltage_point_name: str = None,
        ramp_duration: Union[int, QuaExpression, QuaVariable] = None,
    ) -> None:
        """Add a voltage level to the pulse sequence.
        The voltage level is either identified by its voltage_point_name if added to the voltage_point dict beforehand, or by its level and duration.
        A ramp_duration can be used to ramp to the desired level instead of stepping to it.

        :param level: Desired voltage level of the different gates composing the virtual gate in Volt.
        :param duration: How long the voltage level should be maintained in ns. Must be a multiple of 4ns and either larger than 16ns or 0.
        :param voltage_point_name: Name of the voltage level if added to the list of relevant points in the charge stability map.
        :param ramp_duration: Duration in ns of the ramp if the voltage should be ramped to the desired level instead of stepped. Must be a multiple of 4ns and larger than 16ns.
        """
        self._check_duration(duration)
        if isinstance(ramp_duration, (int, float)):
            if ramp_duration == 0:
                ramp_duration = None
        self._check_duration(ramp_duration)
        if ramp_duration is not None:
            if self.is_QUA(ramp_duration):
                warn(
                    "\nYou are using a QUA variable for the ramp duration, make sure to stay at the final voltage level for more than 52ns or errors/gaps may occur, otherwise use a python variable.",
                    stacklevel=2,
                )

        if level is not None:
            if type(level) is not list or len(level) != len(self._elements):
                raise TypeError(
                    "the provided level must be a list of same length as the number of elements involved in the virtual gate."
                )
            level[:] = [float(x) if isinstance(x, int) else x for x in level]

        if voltage_point_name is not None and duration is None:
            _duration = self._voltage_points[voltage_point_name]["duration"]
        elif duration is not None:
            _duration = duration
        else:
            raise RuntimeError(
                "Either the voltage_point_name or the duration and desired voltage level must be provided."
            )

        for i, gate in enumerate(self._elements):
            if voltage_point_name is not None and level is None:
                voltage_level = self._voltage_points[voltage_point_name]["coordinates"][i]
            elif level is not None:
                voltage_point_name = "unregistered_value"
                voltage_level = level[i]
            else:
                raise RuntimeError(
                    "Either the voltage_point_name or the duration and desired voltage level must be provided."
                )
            # Play a step
            if ramp_duration is None:
                self.average_power[i] += self._update_averaged_power(voltage_level, _duration)

                # Dynamic amplitude change...
                if self.is_QUA(voltage_level) or self.is_QUA(self.current_level[i]):
                    # if dynamic duration --> play step and wait
                    if self.is_QUA(_duration):
                        play(
                            "step"
                            * amp((voltage_level - self.current_level[i]) << self.base_operation[gate]["bit_shift"]),
                            gate,
                        )
                        wait((_duration - 16) >> 2, gate)
                    # if constant duration --> new operation and play(*amp(..))
                    else:
                        if _duration > 0:
                            operation = self._add_op_to_config(
                                gate,
                                "step",
                                amplitude=self.base_operation[gate]["amplitude"],
                                length=_duration,
                            )
                            play(
                                operation
                                * amp(
                                    (voltage_level - self.current_level[i]) << self.base_operation[gate]["bit_shift"]
                                ),
                                gate,
                            )

                # Fixed amplitude but dynamic duration --> new operation and play(duration=..)
                elif isinstance(_duration, (QuaExpression, QuaVariable)):
                    operation = self._add_op_to_config(
                        gate,
                        voltage_point_name,
                        amplitude=voltage_level - self.current_level[i],
                        length=16,
                    )
                    play(operation, gate, duration=_duration >> 2)

                # Fixed amplitude and duration --> new operation and play()
                else:
                    operation = self._add_op_to_config(
                        gate,
                        voltage_point_name,
                        amplitude=voltage_level - self.current_level[i],
                        length=_duration,
                    )
                    play(operation, gate)

            # Play a ramp
            else:
                self._check_amplified_mode(gate)  # Check if the amplified mode is used until the bug is fixed
                self.average_power[i] += self._update_averaged_power(
                    voltage_level, _duration, ramp_duration, self.current_level[i]
                )

                if not self.is_QUA(ramp_duration):
                    ramp_rate = 1 / ramp_duration
                    play(ramp((voltage_level - self.current_level[i]) * ramp_rate), gate, duration=ramp_duration >> 2)
                    if self.is_QUA(_duration) or _duration > 0:
                        wait(_duration >> 2, gate)

                else:
                    ramp_rate = declare(fixed)
                    assign(ramp_rate, (voltage_level - self.current_level[i]) * Math.div(1, ramp_duration))
                    play(ramp(ramp_rate), gate, duration=ramp_duration >> 2)
                    if self.is_QUA(_duration):
                        wait((_duration >> 2) - 9, gate)
                    else:
                        if _duration > 0:
                            wait(_duration >> 2, gate)

            self.current_level[i] = voltage_level

    def add_compensation_pulse(self, max_amplitude: float = 0.49, **kwargs) -> None:
        """Add a compensation pulse of the specified amplitude whose duration is derived automatically from the previous operations and the maximum amplitude allowed.
        Note that the derivation of the compensation pulse parameters in real-time may add a gap up to 300ns before playing the pulse, but the voltage will be maintained.
        :param max_amplitude: Maximum amplitude allowed for the compensation pulse in V. Default is 0.49V.
        """
        duration = kwargs.get("duration", None)

        if duration is not None:
            warn(
                "The duration argument is deprecated and will be ignored in future versions. From qualang-tools 0.20, the compensation pulse duration is derived automatically based on the maximum amplitude allowed.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._check_duration(duration)

        for i, gate in enumerate(self._elements):
            if not self.is_QUA(self.average_power[i]):
                if duration is None:
                    # Exact duration of the compensation pulse
                    comp_duration = max(np.abs(0.0009765625 * self.average_power[i] / max_amplitude), 16)
                    # Duration as an integer multiple of 4ns
                    duration_4ns = max((int(np.ceil(comp_duration)) // 4 + 1) * 4, 48)
                    # Corrected amplitude to account for the duration casting to integer
                    amplitude = -np.sign(self.average_power[i]) * max_amplitude * comp_duration / duration_4ns
                    # Apply the compensation pulse as a ramp to circumvent the max amplitude limit.
                    ramp_rate = (amplitude - self.current_level[i]) / 16
                    play(ramp(ramp_rate), gate, duration=4)
                    wait((duration_4ns - 16) // 4, gate)

                else:
                    amplitude = -0.0009765625 * self.average_power[i] / duration
                    operation = self._add_op_to_config(
                        gate, "compensation", amplitude=amplitude - self.current_level[i], length=duration
                    )
                    play(operation, gate)
            else:
                if duration is None:
                    with if_(Math.abs(self.average_power[i]) > (self._voltage_tolerance * 1024)):
                        eval_average_power = declare(int)
                        comp_duration = declare(int)
                        duration_4ns = declare(int)
                        duration_4ns_pow2 = declare(int)
                        duration_4ns_pow2_cur = declare(int)
                        amplitude = declare(fixed)
                        # Exact duration of the compensation pulse
                        # take into account a gap of 110ns for the derivation of the compensation pulse
                        assign(
                            eval_average_power,
                            self.average_power[i] + Cast.mul_int_by_fixed(96 * 1024, self.current_level[i]),
                        )
                        assign(
                            comp_duration,
                            Cast.mul_int_by_fixed(Math.abs(eval_average_power), 0.0009765625 / max_amplitude),
                        )
                        # Ensure that it is larger than 16ns
                        with if_(comp_duration < 16):
                            assign(comp_duration, 16)
                        # Duration as an integer multiple of 4ns
                        assign(duration_4ns, (comp_duration >> 2) << 2)
                        # Take the closest power of 2 for reducing gaps + 1 to avoid sending too much power
                        assign(duration_4ns_pow2, 29 + Math.msb(duration_4ns))
                        # Get the actual compensation pulse duration
                        assign(duration_4ns_pow2_cur, 1 << duration_4ns_pow2)
                        # Corrected amplitude to account for the actual duration with respect to the exact one
                        with if_(eval_average_power > (self._voltage_tolerance * 1024)):
                            assign(amplitude, -Cast.mul_fixed_by_int(max_amplitude >> duration_4ns_pow2, comp_duration))
                        with elif_(eval_average_power < -(self._voltage_tolerance * 1024)):
                            assign(amplitude, Cast.mul_fixed_by_int(max_amplitude >> duration_4ns_pow2, comp_duration))
                        # Apply the compensation pulse as a ramp to circumvent the max amplitude limit.
                        ramp_rate = declare(fixed)
                        assign(ramp_rate, (amplitude - self.current_level[i]) >> 4)
                        play(ramp(ramp_rate), gate, duration=4)
                        wait((duration_4ns_pow2_cur - 16) >> 2, gate)
                else:
                    operation = self._add_op_to_config(
                        gate, "compensation", amplitude=self.base_operation[gate]["amplitude"], length=duration
                    )
                    amplitude = declare(fixed)
                    eval_average_power = declare(int)
                    assign(eval_average_power, self.average_power[i])
                    # 0.0009765625 = 1/1024 to compensate for the 1024 factor in the derivation of self.average_power
                    # The calculation is done in two steps to avoid rounding errors
                    assign(amplitude, -Cast.mul_fixed_by_int(0.01 / duration, eval_average_power))
                    assign(amplitude, amplitude * 0.09765625)
                    play(
                        operation * amp((amplitude - self.current_level[i]) << self.base_operation[gate]["bit_shift"]),
                        gate,
                    )

            self.current_level[i] = amplitude

    def ramp_to_zero(self, duration: int = None):
        """Ramp all the gate voltages down to zero Volt and reset the averaged voltage derived for defining the compensation pulse.

        :param duration: How long will it take for the voltage to ramp down to 0V in clock cycles (4ns). If not
            provided, the default pulse duration defined in the configuration will be used.
        """
        for i, gate in enumerate(self._elements):
            ramp_to_zero(gate, duration)
            self.current_level[i] = 0.0
            self.average_power[i] = 0
        if self._expression is not None:
            assign(self._expression, 0)
        if self._expression2 is not None:
            assign(self._expression2, 0)

    def add_points(self, name: str, coordinates: list, duration: int) -> None:
        """Register a relevant voltage point.

        :param name: Name of the voltage point.
        :param coordinates: Voltage value of each gate involved in the virtual gate in V.
        :param duration: How long should the voltages be maintained at this level in ns. Must be larger than 16ns and a multiple of 4ns.
        """
        self._voltage_points[name] = {}
        self._voltage_points[name]["coordinates"] = [float(x) for x in coordinates]
        self._voltage_points[name]["duration"] = duration



import abc
from typing import Any, Dict, List, Optional


class ConfigBuilderClass(abc.ABC):
    def __init__(self, **kwargs) -> None:
        super().__init__()


class ConfigBuilderElement(ConfigBuilderClass):
    def __init__(self, name, **kwargs) -> None:
        super(ConfigBuilderElement, self).__init__()
        self.name = name


class Port(ConfigBuilderClass):
    def __init__(self, controller: str, port_id: int, shareable: bool = False) -> None:
        """
        Base class for controller ports.
        """
        super(Port, self).__init__()
        self.controller = controller
        self.port_id = port_id
        self.shareable = shareable

    @property
    def info(self):
        return self.controller, self.port_id

    def __eq__(self, other) -> bool:
        return self.info == other.info and type(self) is type(other)

    def __str__(self):
        cont, port_id = self.info
        return type(self).__name__ + "(" + str(cont) + " , " + str(port_id) + ")"


class AnalogInputPort(Port):
    def __init__(
        self,
        controller: str,
        port_id: int,
        offset: float = 0,
        gain_db: Optional[float] = None,
    ):
        """Analog input port of the controller
        :param controller: name of the controller
        :type controller: str
        :param port_id: port id
        :type port_id: int
        :param offset: offset voltage
        :type offset: float
        :param gain_db: voltage gain in db  (-12dB to 20dB in steps of 1dB)
        :type gain_db: float
        """
        super(AnalogInputPort, self).__init__(controller, port_id)
        self.offset = offset
        self.gain_db = gain_db


class AnalogOutputFilter(ConfigBuilderClass):
    def __init__(self, feedback: List[float], feedforward: List[float]):
        """A filter applied to the analog output ports

        :param feedback: feedback taps for the output filter
        :type feedback: List[float]
        :param feedforward: feedforward taps for the output filter
        :type feedforward: List[float]
        """
        super(AnalogOutputFilter, self).__init__()
        self.feedback = feedback
        self.feedforward = feedforward


class AnalogOutputPort(Port):
    def __init__(
        self,
        controller: str,
        port_id: int,
        offset: float = 0,
        delay: Optional[float] = None,
        filter: Optional[AnalogOutputFilter] = None,
        crosstalk: Optional[dict] = None,
    ):
        """Analog output port of the controller.
        :param controller: name of the controller
        :type controller: str
        :param port_id: port id
        :type port_id: int
        :param offset: offset voltage
        :type offset: float
        :param filter: analog output filter
        :type filter: AnalogOutputPort
        :param crosstalk: crosstalk terms between any ports in the same controller
        :type crosstalk: dict
        """
        super(AnalogOutputPort, self).__init__(controller, port_id)
        self.offset = offset
        self.delay = delay
        self.filter = filter
        self.crosstalk = crosstalk


class DigitalInputPort(Port):
    def __init__(
        self,
        controller: str,
        port_id: int,
        polarity: Optional[float] = None,
        window: Optional[float] = None,
        threshold: Optional[float] = None,
    ):
        """Digital input port of the controller.
        :param controller: name of the controller
        :type controller: str
        :param port_id: port id
        :type port_id: int
        :param polarity: whether it is triggered when 'HIGH'(rising edge) or 'LOW' (falling edge)
        :type polarity: str
        :param window: minimal time between pulses in ns
        :type window: float
        :param threshold: voltage threshold
        :type threshold: float
        """
        super(DigitalInputPort, self).__init__(controller, port_id)
        self.polarity = polarity
        self.window = window
        self.threshold = threshold


class DigitalOutputPort(Port):
    def __init__(self, controller: str, port_id: int):
        """A digital output port of the controller.
        :param controller: name of the controller
        :type controller: str
        :param port_id: port id
        :type port_id: int
        """
        super(DigitalOutputPort, self).__init__(controller, port_id)


class Waveform(ConfigBuilderElement):
    def __init__(self, name: str, dictionary: Dict[str, Any]):
        """A waveform played by the QOP

        :param name: name of the played WF
        :type name: str
        :param dictionary: parameters to initialize a WF
        :type dictionary: Dict[str, Any]
        """
        super(Waveform, self).__init__(name)
        self.dict = dictionary


class Pulse(ConfigBuilderElement):
    def __init__(
        self,
        name: str,
        wfs: List[Waveform],
        operation: str,
        length: int,
        digital_marker: Optional[float] = None,
    ):
        """A an analog pulse

        :param name: Name for this pulse
        :type name: str
        :param wfs: Name of the waveform
        :type wfs: List[Waveform]
        :param operation:
        :type operation: str
        :param length:
        :type length: int
        """
        super(Pulse, self).__init__(name)
        assert len(wfs) <= 2
        self.wfs = wfs
        self.operation = operation
        self.length = length
        self.digital_marker = digital_marker

        self.dict = dict()
        self.dict["operation"] = operation
        self.dict["length"] = length
        self.dict["waveforms"] = dict()
        if len(wfs) == 2:
            self.dict["waveforms"]["I"] = wfs[0].name
            self.dict["waveforms"]["Q"] = wfs[1].name
        elif len(wfs) == 1:
            self.dict["waveforms"]["single"] = wfs[0].name

    @property
    def waveform_names(self):
        return [wf.name for wf in self.wfs]


class Operation(ConfigBuilderClass):
    def __init__(self, pulse: Pulse, name: Optional[str] = ""):
        """An operation used in the QUA program (a pulse with a given name).
        :param Pulse: a pulse
        :type Pulse: Pulse
        :param name: operation name (defaults to the pulse name)
        :type name: str
        """
        super(Operation, self).__init__()
        self.pulse = pulse
        self.name = pulse.name if name == "" else name


class IntegrationWeights(ConfigBuilderElement):
    def __init__(self, name: str, cosine: List, sine: List):
        """A vector of integration weights used in integration and demodulation

        :param name: name for the weights vector
        :type name: str
        :param cosine: a list of weights to be used with the cosine demodulation element, specified at 1 sample per nanosecond
        :type cosine: List
        :param sine: a list of weights to be used with the sine demodulation element, specified at 1 sample per nanosecond
        :type sine: List
        """
        super(IntegrationWeights, self).__init__(name)
        self.dict = dict()
        self.dict["cosine"] = cosine
        self.dict["sine"] = sine

    @property
    def cosine(self):
        return self.dict["cosine"]

    @cosine.setter
    def cosine(self, w: List[float]):
        self.dict["cosine"] = w

    @property
    def sine(self):
        return self.dict["sine"]

    @sine.setter
    def sine(self, w: List[float]):
        self.dict["sine"] = w


class Weights(ConfigBuilderClass):
    def __init__(self, weights: IntegrationWeights, name: Optional[str] = ""):
        """Integration weights with a given name.
        :param weights: Integation weights used in demodulation or integration.
        :type weights: IntegrationWeights
        :param name: name of the integration weights (this name is used in the measure statement)
        :type name: str
        """
        super(Weights, self).__init__()
        self.name = weights.name if name == "" else name
        self.weights = weights


class DigitalSample(ConfigBuilderClass):
    def __init__(self, state: int, duration: int):
        """A sample to describe digital waveform

        :param state: state of the digital output either 0 or 1
        :type state: int
        :param duration: duration of the output in nanoseconds
        :type duration: int
        """
        super(DigitalSample, self).__init__()
        assert state == 0 or state == 1
        self.state = state
        self.duration = duration


class Matrix2x2(ConfigBuilderClass):
    def __init__(self, data: List[List[float]]):
        """A correction matrix for the IQ mixer

        :param data: 2x2 matrix provided as a list containing a list of rows
        :type data: List[List[float]]
        """
        super(Matrix2x2, self).__init__()
        assert len(data) == 2
        assert len(data[0]) == 2
        assert len(data[1]) == 2
        self.data = data


class MixerData(ConfigBuilderClass):
    def __init__(
        self,
        intermediate_frequency: int,
        lo_frequency: int,
        correction: Matrix2x2,
    ):
        """Mixer data

        :param intermediate_frequency: intermediate frequency
        :type intermediate_frequency: int
        :param lo_frequency: frequency of the local oscillator
        :type lo_frequency: int
        :param correction: correction matrix of the mixer
        :type correction: Matrix2x2
        """
        super().__init__()
        self.intermediate_frequency = intermediate_frequency
        self.lo_frequency = lo_frequency
        self.correction = correction



from typing import Dict, Any, Generator, List, Tuple, Optional


def iterate_nested_dict(
    d: Dict[str, Any], parent_keys: Optional[List[str]] = None
) -> Generator[Tuple[List[str], Any], None, None]:
    """Iterate over a nested dictionary (dicts only).

    :param d: The dictionary to iterate over
    :param parent_keys: The keys of the parent dictionary. Used for recursion
    :return: A generator that yields a tuple of the keys and the value
    """
    if parent_keys is None:
        parent_keys = []
    for k, v in d.items():
        keys = parent_keys + [k]
        yield keys, v
        if isinstance(v, dict):
            yield from iterate_nested_dict(v, parent_keys=keys)


def iterate_nested_dict_with_lists(
    d: Any, parent_keys: Optional[List[str]] = None
) -> Generator[Tuple[List[str], Any], None, None]:
    """Iterate over a nested dictionary and lists.

    :param d: The dictionary or list to iterate over
    :param parent_keys: The keys of the parent dictionary. Used for recursion
    :return: A generator that yields a tuple of the keys and the value
    """
    if parent_keys is None:
        parent_keys = []
    if isinstance(d, dict):
        for k, v in d.items():
            keys = parent_keys + [k]
            yield keys, v
            yield from iterate_nested_dict_with_lists(v, parent_keys=keys)
    elif isinstance(d, list):
        for idx, item in enumerate(d):
            keys = parent_keys + [str(idx)]
            yield keys, item
            yield from iterate_nested_dict_with_lists(item, parent_keys=keys)


def update_nested_dict(d: dict, keys: list, value: Any) -> None:
    """Update a nested dictionary or list with a new value

    :param d: The dictionary or list to update
    :param keys: The keys (or string indices) to the value to update
    :param value: The new value to set
    """
    sub = d
    for key in keys[:-1]:
        if isinstance(sub, list):
            idx = int(key)
            sub = sub[idx]
        else:
            sub = sub[key]
    last_key = keys[-1]
    if isinstance(sub, list):
        sub[int(last_key)] = value
    else:
        sub[last_key] = value


def copy_nested_dict(d: dict) -> dict:
    """Copy a nested dictionary, but don't make copies of the values

    This function will copy a nested dictionary, but will not make copies of the values. This is useful if copying the
    values may be an expensive operation (e.g. large arrays).
    If you also want to make copies of the values, use `copy.deepcopy`

    :param d: The dictionary to copy
    :return: A new dictionary with the same structure as `d`, but with the same values
    """
    new_dict = {}
    for key, val in d.items():
        if isinstance(val, dict):
            new_dict[key] = copy_nested_dict(val)
        else:
            new_dict[key] = val
    return new_dict



from typing import List, Union

from .instrument_channel import (
    InstrumentChannelOctaveInput,
    InstrumentChannelOctaveOutput,
    InstrumentChannelOctaveDigitalInput,
    InstrumentChannelExternalMixerInput,
    InstrumentChannelExternalMixerOutput,
    InstrumentChannelExternalMixerDigitalInput,
)
from .instrument_channels import *
from .instrument_pulsers import *
from .constants import *


class Instruments:
    """
    Class to add the static information about which QM instruments will be used in an experimental setup.
    Upon adding an instrument with one of the `add_` methods, its available channels will be enumerated and added
    individually to a public data structure of "available" channels.
    Later, these can be moved into a public data structure of "used" channels during allocation.
    """

    def __init__(self):
        self.used_channels = InstrumentChannels()
        self.available_channels = InstrumentChannels()
        self.used_pulsers = InstrumentPulsers()
        self.available_pulsers = InstrumentPulsers()

    def add_external_mixer(self, indices: Union[List[int], int]):
        """
        Add an external mixer, which is defined abstractly as a combined IQ-upconverter and
        IQ-downconverter.

        `indices` (List[int] | int): Can be one or more indices for one or more external mixers.
        """
        if isinstance(indices, int):
            indices = [indices]

        for index in indices:
            channel = InstrumentChannelExternalMixerInput(con=index, port=1)
            self.available_channels.add(channel)

            channel = InstrumentChannelExternalMixerOutput(con=index, port=1)
            self.available_channels.add(channel)

            channel = InstrumentChannelExternalMixerDigitalInput(con=index, port=1)
            self.available_channels.add(channel)

    def add_octave(self, indices: Union[List[int], int]):
        """
        Adds N octaves to the available instrument channels with the specified indices.

        This method handles both single integer indices and lists of indices.
        It creates and adds three types of channels: input, output, and digital input,
        for each octave index provided.

        Args:
            indices (Union[List[int], int]): Either a single integer or a list of integers
                representing the indices of the octaves to be added.
        """
        if isinstance(indices, int):
            indices = [indices]

        for index in indices:
            for port in range(1, NUM_OCTAVE_INPUT_PORTS + 1):
                channel = InstrumentChannelOctaveInput(con=index, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_OCTAVE_OUTPUT_PORTS + 1):
                channel = InstrumentChannelOctaveOutput(con=index, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_OCTAVE_DIGITAL_INPUT_PORTS + 1):
                channel = InstrumentChannelOctaveDigitalInput(con=index, port=port)
                self.available_channels.add(channel)

    def add_lf_fem(self, controller: int, slots: Union[List[int], int]):
        """
        Adds Low-Frequency Front End Module (LF-FEM) channels to the available channels of the
        instrument, labelled according to the specified `controller` index and `slot` indices.

        This method configures the LF-FEM input channels, output channels, and digital output
        channels for the given controller and for N LF-FEMs in the provided `slots`. If a single
        slot is provided as an integer, it will be converted into a list of slots internally.

        Args:
            controller (int): The controller ID associated with the LF FEM channels.
            slots (Union[List[int], int]): One or more slots for which the LF-FEMs reside in the
                OPX1000 chassis
        """
        if isinstance(slots, int):
            slots = [slots]

        for slot in slots:
            for port in range(1, NUM_LF_FEM_INPUT_PORTS + 1):
                channel = InstrumentChannelLfFemInput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_LF_FEM_OUTPUT_PORTS + 1):
                channel = InstrumentChannelLfFemOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_LF_FEM_DIGITAL_OUTPUT_PORTS + 1):
                channel = InstrumentChannelLfFemDigitalOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for idx in range(NUM_THREADS_PER_FEM):
                pulser = Pulser(controller=controller, slot=slot)
                self.available_pulsers.add(pulser)

    def add_mw_fem(self, controller: int, slots: Union[List[int], int]):
        """
        Adds microwave front-end module (MW-FEM) input, output, and digital output channels
        to the available channels for the given controller and specified slots.

        This method dynamically creates and registers MW-FEM channels based on the provided
        controller and slots. The slots can be specified as a single integer or as a list
        of integers. For each slot, the input, output, and digital output channels are
        instantiated and added to the existing pool of available channels.

        Args:
            controller (int): The identifier for the controller to which the MW FEMs should
                be added.
            slots (Union[List[int], int]): One or more slot numbers where the MW FEMs are
                located in the OPX1000 chassis. If given a single integer, it will be treated
                as a list with one slot.
        """
        if isinstance(slots, int):
            slots = [slots]

        for slot in slots:
            for port in range(1, NUM_MW_FEM_INPUT_PORTS + 1):
                channel = InstrumentChannelMwFemInput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_MW_FEM_OUTPUT_PORTS + 1):
                channel = InstrumentChannelMwFemOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_MW_FEM_OUTPUT_PORTS + 1):
                channel = InstrumentChannelMwFemDigitalOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for idx in range(NUM_THREADS_PER_FEM):
                pulser = Pulser(controller=controller, slot=slot)
                self.available_pulsers.add(pulser)

    def add_opx_plus(self, controllers: Union[List[int], int]):
        """
        Adds OPX+ controllers and their associated channels to the available pool of
        channels. The method processes input, output, and digital output ports for
        each specified controller. Digital output ports are added in a specific order
        (odd-numbered ports first, followed by even-numbered ports) for convenience
        and potential compatibility with octave connectivity.

        Note that digital output ports are added in the following order: 1, 3, 5, 7, 9,
        2, 4, 6, 8, 10. This ordering corresponds to the physical order of the channels,
        starting from left-to-right, then top-to-bottom.

        Args:
            controllers (Union[List[int], int]): A list of controller IDs or a single
                controller ID. The controllers represent the OPX+ devices to be
                configured and their ports to be allocated.
        """
        if isinstance(controllers, int):
            controllers = [controllers]

        for controller in controllers:
            for port in range(1, NUM_OPX_PLUS_INPUT_PORTS + 1):
                channel = InstrumentChannelOpxPlusInput(con=controller, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_OPX_PLUS_OUTPUT_PORTS + 1):
                channel = InstrumentChannelOpxPlusOutput(con=controller, port=port)
                self.available_channels.add(channel)

            # add odd first for octave connectivity (top row is more convenient)
            for port in range(1, NUM_OPX_PLUS_DIGITAL_OUTPUT_PORTS + 1, 2):
                channel = InstrumentChannelOpxPlusDigitalOutput(con=controller, port=port)
                self.available_channels.add(channel)

            # then add evens
            for port in range(2, NUM_OPX_PLUS_DIGITAL_OUTPUT_PORTS + 1, 2):
                channel = InstrumentChannelOpxPlusDigitalOutput(con=controller, port=port)
                self.available_channels.add(channel)

            # Finally add the pulsers for this controller.
            for idx in range(NUM_THREADS_PER_OPX_PLUS):
                pulser = Pulser(controller=controller, slot=None)
                self.available_pulsers.add(pulser)



import matplotlib.pyplot as plt
import numpy as np


def _round_to_fixed_point_accuracy(x, accuracy=2**-15):
    return np.round(x / accuracy) * accuracy


def convert_integration_weights(integration_weights, N=100, accuracy=2**-15, plot=False):
    """
    Converts a list of integration weights, in which each sample corresponds to a clock cycle (4ns), to a list
    of tuples with the format (weight, time_to_integrate_in_ns).
    Can be used to convert between the old format (up to QOP 1.10) to the new format introduced in QOP 1.20.

    :param integration_weights: A list of integration weights.
    :param N:   Maximum number of tuples to return. The algorithm will first create a list of tuples, and then if it is
                too long, it will run :func:`compress_integration_weights` on them.
    :param accuracy:    The accuracy at which to calculate the integration weights. Default is 2^-15, which is
                        the accuracy at which the OPX operates for the integration weights.
    :param plot: If true, plots the integration weights before and after the conversion.
    :type integration_weights: list[float]
    :type N: int
    :type accuracy: float
    :type plot: bool
    :return: List of tuples representing the integration weights
    """
    integration_weights = np.array(integration_weights)
    integration_weights = _round_to_fixed_point_accuracy(integration_weights, accuracy)
    changes_indices = np.where(np.abs(np.diff(integration_weights)) > 0)[0].tolist()
    prev_index = -1
    new_integration_weights = []
    for curr_index in changes_indices + [len(integration_weights) - 1]:
        constant_part = (
            integration_weights[curr_index].tolist(),
            round(4 * (curr_index - prev_index)),
        )
        new_integration_weights.append(constant_part)
        prev_index = curr_index

    new_integration_weights = compress_integration_weights(new_integration_weights, N=N, plot=plot)

    return new_integration_weights


def compress_integration_weights(integration_weights, N=100, plot=False):
    """
    Compresses the list of tuples with the format (weight, time_to_integrate_in_ns) to one with length < N.
    Works by iteratively finding the nearest integration weights and combining them with a weighted average.

    :param integration_weights: The integration_weights to be compressed.
    :param N: The maximum list length required.
    :param plot: If true, plots the integration weights before and after the conversion.
    :return: The compressed list of tuples representing the integration weights.
    """
    integration_weights_before = integration_weights
    integration_weights = np.array(integration_weights)
    while len(integration_weights) > N:
        diffs = np.abs(np.diff(integration_weights, axis=0)[:, 0])
        min_diff = np.min(diffs)
        min_diff_indices = np.where(diffs == min_diff)[0][0]
        times1 = integration_weights[min_diff_indices, 1]
        times2 = integration_weights[min_diff_indices + 1, 1]
        weights1 = integration_weights[min_diff_indices, 0]
        weights2 = integration_weights[min_diff_indices + 1, 0]
        integration_weights[min_diff_indices, 0] = (weights1 * times1 + weights2 * times2) / (times1 + times2)
        integration_weights[min_diff_indices, 1] = times1 + times2
        integration_weights = np.delete(integration_weights, min_diff_indices + 1, 0)
    integration_weights = list(
        zip(
            integration_weights.T[0].tolist(),
            integration_weights.T[1].astype(int).tolist(),
        )
    )
    if plot:
        plt.figure()
        plot_integration_weights(integration_weights_before, label="Original")
        plot_integration_weights(integration_weights, label="Compressed")
        plt.show()
    return integration_weights


def plot_integration_weights(integration_weights, label=None):
    """
    Plot the integration weights in units of ns, receives the integration weights in both formats

    :param integration_weights: The integration_weights to be plotted.
    :param str label: The label of the integration_weights.
    """

    if isinstance(integration_weights[0], tuple):
        a = [[i[0]] * i[1] for i in integration_weights]
        unpacked_weights = sum(a, start=[])
    elif isinstance(integration_weights[0], float):
        a = [[i] * 4 for i in integration_weights]
        unpacked_weights = sum(a, start=[])
    else:
        raise Exception("Unknown input")

    plt.plot(unpacked_weights, label=label)
    if label is not None:
        plt.legend()



# Octave Gain in dB
OCTAVE_GAIN_MIN = -20
OCTAVE_GAIN_MAX = 20
OCTAVE_GAIN_STEP = 0.5
# OPX+ waveform amplitude in V
OPX_AMPLITUDE_MIN = -0.5
OPX_AMPLITUDE_MAX = 0.5 - 2**-28
# OPX1000 MW-FEM waveform amplitude
OPX1000_MW_AMPLITUDE_MIN = -1.0
OPX1000_MW_AMPLITUDE_MAX = 1.0
# OPX1000 MW-FEM full scale power in dBm
OPX1000_MW_POWER_MIN = -11
OPX1000_MW_POWER_MAX = 16
OPX1000_MW_POWER_STEP = 3
# OPX1000 MW-FEM bands
OPX1000_MW_BANDS = {1: (50e6, 5.5e9), 2: (4.5e9, 7.5e9), 3: (6.5e9, 10.5e9)}



from typing import Dict, List, Union

from .channel_spec import ChannelSpec
from .element import Element, ElementId, QubitReference, QubitPairReference
from .types import QubitsType, QubitPairsType
from .wiring_spec import WiringSpec, WiringFrequency, WiringIOType, WiringLineType


class ConnectivityBase:
    """
    The base class for managing quantum element wiring configurations.

    This class serves as the foundation for storing and managing placeholders for
    quantum elements (such as qubits and qubit pairs) and their associated wiring
    specifications. It tracks each element's unique ID and the specific wiring
    configurations required to integrate those elements into a quantum setup.

    The wiring specifications include details such as:
    - Frequency at which the element is driven (e.g., RF or DC).
    - Input/output requirements for the element (e.g., input, output, or both).
    - The specific line type (e.g., resonator, drive, flux) for the element.
    - Any constraints on the channels (e.g., frequency, module location, etc.).
    - The associated high-level quantum components (e.g., qubits, resonators, etc.).

    Attributes:
        elements (Dict[ElementId, Element]): A dictionary mapping each element's
                                              unique ID to its corresponding `Element` object.
        specs (List[WiringSpec]): A list of all the wiring specifications (placeholders)
                                  added for the quantum elements.
    """

    def __init__(self):
        """
        Initializes a new `ConnectivityBase` instance with empty storage for elements
        and wiring specifications.

        This constructor sets up two primary containers:
        - `self.elements`: A dictionary that holds the quantum elements (e.g., qubits).
        - `self.specs`: A list that stores wiring specifications that detail how
                        the quantum elements are connected and configured.
        """
        self.elements: Dict[ElementId, Element] = {}
        self.specs: List[WiringSpec] = []

    def add_wiring_spec(
        self,
        frequency: WiringFrequency,
        io_type: WiringIOType,
        line_type: Union[WiringLineType, str],
        triggered: bool,
        constraints: ChannelSpec,
        elements: List[Element],
        shared_line: bool = False,
    ):
        """
        Adds wiring specifications (placeholders) for quantum elements.

        This method allows the addition of one or more wiring specifications for
        a set of quantum elements. Each specification defines the configuration of
        how the element will be wired, including frequency, input/output lines,
        line type, triggering conditions, and channel constraints. The specifications
        serve as placeholders, and no channels are actually allocated at this stage.

        Args:
            frequency (WiringFrequency): The frequency at which the element will be driven (e.g., RF or DC).
            io_type (WiringIOType): The input/output configuration for the element (e.g., input, output, or both).
            line_type (Union[WiringLineType, str]): The type of line for the wiring (e.g., resonator, drive, flux).
            triggered (bool): Whether the wiring is triggered (e.g., by an external event).
            constraints (ChannelSpec): Any channel constraints that should be applied (e.g., frequency domain,
                                       slot availability, etc.).
            elements (List[Element]): The quantum elements to which the wiring specifications apply.
            shared_line (bool, optional): Whether the wiring specification should apply to a shared line
                                          (defaults to False).

        Returns:
            List[WiringSpec]: A list of the generated wiring specifications for the provided elements.
        """
        specs = []
        for element in elements:
            if element.id not in self.elements:
                self._add_element(element)

        if shared_line:
            spec = WiringSpec(frequency, io_type, line_type, triggered, constraints, elements)
            specs.append(spec)
        else:
            for element in elements:
                spec = WiringSpec(frequency, io_type, line_type, triggered, constraints, element)
                specs.append(spec)
        self.specs.extend(specs)

        return specs

    def _make_qubit_elements(self, qubits: QubitsType):
        """
        Creates `Element` objects for a list of qubits.

        This method constructs `Element` objects for each qubit reference in the provided `qubits` list
        and adds them to the internal `self.elements` dictionary. If the element for a qubit already exists,
        it will not be recreated.

        Args:
            qubits (QubitsType): A list or a single qubit reference to generate element objects for.

        Returns:
            List[Element]: A list of `Element` objects created or retrieved for the specified qubits.
        """
        if not isinstance(qubits, list):
            qubits = [qubits]

        elements = []
        for qubit in qubits:
            id = QubitReference(qubit)
            if id not in self.elements:
                self.elements[id] = Element(id)
            elements.append(self.elements[id])

        return elements

    def _make_qubit_pair_elements(self, qubit_pairs: QubitPairsType):
        """
        Creates `Element` objects for a list of qubit pairs.

        This method constructs `Element` objects for each qubit pair reference in the provided `qubit_pairs` list
        and adds them to the internal `self.elements` dictionary. If the element for a qubit pair already exists,
        it will not be recreated.

        Args:
            qubit_pairs (QubitPairsType): A list or a single qubit pair reference to generate element objects for.

        Returns:
            List[Element]: A list of `Element` objects created or retrieved for the specified qubit pairs.
        """
        if not isinstance(qubit_pairs, list):
            qubit_pairs = [qubit_pairs]

        elements = []
        for qubit_pair in qubit_pairs:
            id = QubitPairReference(*qubit_pair)
            if id not in self.elements:
                self.elements[id] = Element(id)
            elements.append(self.elements[id])

        return elements

    def _add_elements(self, elements: List[Element]):
        """
        Adds multiple `Element` objects to the internal storage.

        This method adds a list of `Element` objects to the internal `self.elements` dictionary.

        Args:
            elements (List[Element]): A list of `Element` objects to add to the internal storage.
        """
        for element in elements:
            self._add_element(element)

    def _add_element(self, element: Element):
        """
        Adds a single `Element` object to the internal storage.

        This method adds a single `Element` object to the internal `self.elements` dictionary, using
        the element's unique ID as the key.

        Args:
            element (Element): The `Element` object to add to the internal storage.
        """
        self.elements[element.id] = element



from dataclasses import dataclass
from typing import Callable

from qualang_tools.bakery.bakery import baking, Baking
import numpy as np

rnd_gate_list = ["sx", "sy", "sw", "id"]

op_func = Callable[[Baking], None]


@dataclass
class XEBOpsSingleQubit:
    """
    A class for representing the XEB operations. Each callable corresponds to a function with a baking object as argument,
    and performs the operations in the body of the function.

    Operations defined in this class correspond to single qubit operations
    """

    id: op_func
    sx: op_func
    sy: op_func
    sw: op_func


class XEB:
    def __init__(
        self,
        config: dict,
        m_max: int,
        q1_ops: XEBOpsSingleQubit,
        q2_ops: XEBOpsSingleQubit,
        two_qubit_op: op_func,
        align_op: op_func,
    ):
        """
        Class instance for cross-entropy benchmarking sequence generation.

        :param config: Configuration file
        :param m_max: Maximum length of XEB sequence
        :param q1_ops: operations dataclass which maps single qubit operations to series of operations for qubit 1
        :param q2_ops: operations dataclass which maps single qubit operations to series of operations for qubit 2
        :param two_qubit_op: a callable with baking object as argument that performs the two qubit operation
        :param align_op: a callable with baking object as argument that performs an align operation on all quantum elements

        """
        self.config = config
        self.m_max = m_max
        self.duration_tracker = [0] * m_max
        self.operations_list = {qe: [] for qe in ["q1", "q2"]}
        self.baked_sequence = self._generate_xeb_sequence(q1_ops, q2_ops, two_qubit_op, align_op)

    def _generate_xeb_sequence(self, q1_ops, q2_ops, two_qubit_op, align_op):
        rand_seq1 = np.random.randint(3, size=self.m_max)
        rand_seq2 = np.random.randint(3, size=self.m_max)

        with baking(self.config) as b_2_qubit:
            two_qubit_op(b_2_qubit)

        b_2_qubit_len = get_total_len(b_2_qubit)

        b_q1 = []
        q1_len = 0
        for op in rnd_gate_list:
            with baking(self.config) as btemp:
                getattr(q1_ops, op)(btemp)
            b_q1.append(btemp)
            q1_len = max(q1_len, get_total_len(btemp))

        b_q2 = []
        q2_len = 0
        for op in rnd_gate_list:
            with baking(self.config) as btemp:
                getattr(q2_ops, op)(btemp)
            b_q2.append(btemp)
            q2_len = max(q2_len, get_total_len(btemp))

        with baking(self.config) as b_main:
            i = 0
            for rnd1, rnd2 in zip(rand_seq1, rand_seq2):
                align_op(b_main)
                play_all_ops(b_main, b_q1[rnd1])
                play_all_ops(b_main, b_q2[rnd2])
                align_op(b_main)
                play_all_ops(b_main, b_2_qubit)
                align_op(b_2_qubit)
                self.operations_list["q1"].append(rnd_gate_list[rnd1])
                self.operations_list["q2"].append(rnd_gate_list[rnd2])
                self.duration_tracker[i] = max(q1_len, q2_len) + b_2_qubit_len
                i += 1
            play_all_ops(b_main, b_q1[3])
            play_all_ops(b_main, b_q2[3])
        self.duration_tracker = np.cumsum(self.duration_tracker).tolist()
        return b_main


def play_all_ops(current_bake, sub_bake):
    for element in sub_bake.elements:
        current_bake.play(sub_bake.operations[element], element)


def get_total_len(baking_sequence):
    return max(baking_sequence.get_op_length(element) for element in baking_sequence.elements)



from pathlib import Path

from .helpers import copy_nested_dict, iterate_nested_dict, update_nested_dict
from .data_processor import DataProcessor


class XarraySaver(DataProcessor):
    merge_arrays: bool = False
    merged_array_name: str = "xarrays"
    file_format: str = "hdf5"
    nested_separator: str = "."

    def __init__(self, merge_arrays=None, merged_array_name=None, file_format=None):
        if merge_arrays is not None:
            self.merge_arrays = merge_arrays
        if merged_array_name is not None:
            self.merged_array_name = merged_array_name
        if file_format is not None:
            self.file_format = file_format

        self.data_arrays = {}

    @property
    def file_suffix(self) -> str:
        suffixes = {"nc": ".nc", "netcdf": ".nc", "h5": ".h5", "hdf5": ".h5", "zarr": ".zarr"}
        return suffixes[self.file_format.lower()]

    def process(self, data):
        import xarray as xr

        self.data_arrays = {}
        processed_data = copy_nested_dict(data)

        for keys, val in iterate_nested_dict(data):
            if not isinstance(val, xr.Dataset):
                continue

            path = Path(self.nested_separator.join(keys))
            self.data_arrays[path] = val
            if self.merge_arrays:
                merged_array_name = Path(self.merged_array_name).with_suffix(self.file_suffix)
                update_nested_dict(processed_data, keys, f"./{merged_array_name}#{path}")
            else:
                update_nested_dict(processed_data, keys, f"./{path}{self.file_suffix}")
        return processed_data

    def save_merged_netcdf_arrays(self, path: Path, arrays: dict):
        for array_path, array in self.data_arrays.items():
            try:
                array.to_netcdf(path, mode="a", group=str(array_path))
            except ValueError as e:
                raise ValueError(
                    f"Error saving merged array {path}. You may neet to first run `pip install netcdf4`"
                ) from e

    def post_process(self, data_folder: Path):
        if self.file_suffix not in [".nc", ".h5"]:
            raise NotImplementedError(f"File format {self.file_format} is not supported")

        if self.merge_arrays:
            for path, array in self.data_arrays.items():
                merged_path = data_folder / Path(self.merged_array_name).with_suffix(self.file_suffix)
                try:
                    array.to_netcdf(merged_path, mode="a", group=str(path))
                except ValueError as e:
                    raise ValueError(
                        f"Error saving merged array {merged_path}. You may neet to first run `pip install netcdf4`"
                    ) from e
        else:
            for path, array in self.data_arrays.items():
                array.to_netcdf(data_folder / f"{path}{self.file_suffix}")



import dataclasses
import numpy as np
import xarray as xr
from matplotlib import pyplot as plt
from scipy.optimize import curve_fit


@dataclasses.dataclass
class RBResult:
    """
    Class for analyzing and visualizing the results of a Randomized Benchmarking (RB) experiment.

    Attributes:
        circuit_depths (list[int]): List of circuit depths used in the RB experiment.
        num_repeats (int): Number of repeated sequences at each circuit depth.
        num_averages (int): Number of averages for each sequence.
        state (np.ndarray): Measured states from the RB experiment.
    """

    circuit_depths: list[int]
    num_repeats: int
    num_averages: int
    state: np.ndarray

    def __post_init__(self):
        """
        Initializes the xarray Dataset to store the RB experiment data.
        """
        self.data = xr.Dataset(
            data_vars={"state": (["repeat", "circuit_depth", "average"], self.state)},
            coords={
                "repeat": range(self.num_repeats),
                "circuit_depth": self.circuit_depths,
                "average": range(self.num_averages),
            },
        )

    def plot_hist(self, n_cols=3):
        """
        Plots histograms of the N-qubit state distribution at each circuit depth.

        Args:
            n_cols (int): Number of columns in the plot grid. Adjusted if fewer circuit depths are provided.
        """
        if len(self.circuit_depths) < n_cols:
            n_cols = len(self.circuit_depths)
        n_rows = max(int(np.ceil(len(self.circuit_depths) / n_cols)), 1)
        plt.figure()
        for i, circuit_depth in enumerate(self.circuit_depths, start=1):
            ax = plt.subplot(n_rows, n_cols, i)
            self.data.state.sel(circuit_depth=circuit_depth).plot.hist(ax=ax, xticks=range(4))
        plt.tight_layout()

    def plot(self):
        """
        Plots the raw recovery probability decay curve as a function of circuit depth.
        The curve is plotted using the averaged probability and without any fitting.
        """
        recovery_probability = (self.data.state == 0).sum(("repeat", "average")) / (
            self.num_repeats * self.num_averages
        )
        recovery_probability.rename("Recovery Probability").plot.line()

    def plot_with_fidelity(self):
        """
        Plots the RB fidelity as a function of circuit depth, including a fit to an exponential decay model.
        The fitted curve is overlaid with the raw data points, and error bars are included.
        """
        A, alpha, B = self.fit_exponential()
        fidelity = self.get_fidelity(alpha)

        # Compute error bars
        error_bars = (self.data == 0).mean(dim="average").std(dim="repeat").state.data

        plt.figure()
        plt.errorbar(
            self.circuit_depths,
            self.get_decay_curve(),
            yerr=error_bars,
            fmt=".",
            capsize=2,
            elinewidth=0.5,
            color="blue",
            label="Experimental Data",
        )

        circuit_depths_smooth_axis = np.linspace(self.circuit_depths[0], self.circuit_depths[-1], 100)
        plt.plot(
            circuit_depths_smooth_axis,
            rb_decay_curve(np.array(circuit_depths_smooth_axis), A, alpha, B),
            color="red",
            linestyle="--",
            label="Exponential Fit",
        )

        plt.text(
            0.5,
            0.95,
            f"2Q Clifford Fidelity = {fidelity * 100:.2f}%",
            horizontalalignment="center",
            verticalalignment="top",
            fontdict={"fontsize": "large", "fontweight": "bold"},
            transform=plt.gca().transAxes,
        )

        plt.xlabel("Circuit Depth")
        plt.ylabel(r"Probability to recover to $|00\rangle$")
        plt.title("2Q Randomized Benchmarking")
        plt.legend(framealpha=0)
        plt.show()

    def plot_two_qubit_state_distribution(self):
        """
        Plot how the two-qubit state is distributed as a function of circuit-depth on average.
        """
        plt.plot(
            self.circuit_depths,
            (self.data.state == 0).mean(dim="average").mean(dim="repeat").data,
            label=r"$|00\rangle$",
            marker=".",
            color="c",
            linewidth=3,
        )
        plt.plot(
            self.circuit_depths,
            (self.data.state == 1).mean(dim="average").mean(dim="repeat").data,
            label=r"$|01\rangle$",
            marker=".",
            color="b",
            linewidth=1,
        )
        plt.plot(
            self.circuit_depths,
            (self.data.state == 2).mean(dim="average").mean(dim="repeat").data,
            label=r"$|10\rangle$",
            marker=".",
            color="y",
            linewidth=1,
        )
        plt.plot(
            self.circuit_depths,
            (self.data.state == 3).mean(dim="average").mean(dim="repeat").data,
            label=r"$|11\rangle$",
            marker=".",
            color="r",
            linewidth=1,
        )
        plt.axhline(0.25, color="grey", linestyle="--", linewidth=2, label="2Q mixed-state")

        plt.xlabel("Circuit Depth")
        plt.ylabel(r"Probability to recover to a given state")
        plt.title("2Q State Distribution vs. Circuit Depth")
        plt.legend(framealpha=0, title=r"2Q State $\mathbf{|q_cq_t\rangle}$", title_fontproperties={"weight": "bold"})
        plt.show()

    def fit_exponential(self):
        """
        Fits the decay curve of the RB data to an exponential model.

        Returns:
            tuple: Fitted parameters (A, alpha, B) where:
                - A is the amplitude.
                - alpha is the decay constant.
                - B is the offset.
        """
        decay_curve = self.get_decay_curve()

        popt, _ = curve_fit(rb_decay_curve, self.circuit_depths, decay_curve, p0=[0.75, 0.9, 0.25], maxfev=10000)
        A, alpha, B = popt

        return A, alpha, B

    def get_fidelity(self, alpha):
        """
        Calculates the average fidelity per Clifford based on the decay constant.

        Args:
            alpha (float): Decay constant from the exponential fit.

        Returns:
            float: Estimated average fidelity per Clifford.
        """
        n_qubits = 2  # Assuming 2 qubits as per the context
        d = 2**n_qubits
        r = 1 - alpha - (1 - alpha) / d
        fidelity = 1 - r

        return fidelity

    def get_decay_curve(self):
        """
        Calculates the decay curve from the RB data.

        Returns:
            np.ndarray: Decay curve representing the fidelity as a function of circuit depth.
        """
        return (self.data.state == 0).sum(("repeat", "average")) / (self.num_repeats * self.num_averages)


def rb_decay_curve(x, A, alpha, B):
    """
    Exponential decay model for RB fidelity.

    Args:
        x (array-like): Circuit depths.
        A (float): Amplitude of the decay.
        alpha (float): Decay constant.
        B (float): Offset of the curve.

    Returns:
        np.ndarray: Calculated decay curve.
    """
    return A * alpha**x + B


def get_interleaved_gate_fidelity(num_qubits: int, reference_alpha: float, interleaved_alpha: float):
    """
    Calculates the interleaved gate fidelity using the formula from https://arxiv.org/pdf/1210.7011.

    Args:
        num_qubits (int): Number of qubits involved.
        reference_alpha (float): Decay constant from the reference RB experiment.
        interleaved_alpha (float): Decay constant from the interleaved RB experiment.

    Returns:
        float: Estimated interleaved gate fidelity.
    """
    return 1 - ((2**num_qubits - 1) * (1 - interleaved_alpha / reference_alpha) / 2**num_qubits)



import copy
from typing import List, Dict

from qualang_tools.config.components import (
    Controller,
    MeasurePulse,
    Element,
    ConstantWaveform,
    ArbitraryWaveform,
    Mixer,
    Oscillator,
)
from qualang_tools.config.parameters import _is_callable, _is_parametric
from qualang_tools.config.primitive_components import (
    Waveform,
    Pulse,
    IntegrationWeights,
)


class QMConfiguration:
    def __init__(self, controllers: List[Controller] = None, version: int = 1):
        """A class to generate the QM configuration

        :param controllers: A controller or list of controllers used in this configuration, defaults to None
        :type controllers: List[Controller], optional
        :param version: Controller version, defaults to 1
        :type version: int, optional
        """
        self.config = dict()
        self.config["version"] = 1
        self.config["controllers"] = dict()
        for cont in controllers:
            self.config["controllers"][cont.name] = self._call_dict_parameters(cont.dict)
        self.config["elements"] = dict()
        self.config["pulses"] = dict()
        self.config["waveforms"] = dict()
        self.config["digital_waveforms"] = dict()
        self.config["integration_weights"] = dict()
        self.config["mixers"] = dict()
        self.config["oscillators"] = dict()

    def add_waveform(self, wf: Waveform):
        """Add a single waveform to this configuration

        :param wf: a Wavefrom object
        :type wf: Waveform
        """
        self.config["waveforms"][wf.name] = self._call_dict_parameters(wf.dict)

    def _call_dict_parameters(self, dic: Dict):
        if _is_callable(dic):
            if not _is_parametric(dic):
                _dic = copy.deepcopy(dic)
                for k, v in _dic.items():
                    if _is_callable(v):
                        if isinstance(v, dict):
                            _dic[k] = self._call_dict_parameters(v)
                        elif isinstance(v, list):
                            _dic[k] = [e() if _is_callable(e) else e for e in v]
                        else:
                            _dic[k] = v()
                return _dic
        return dic

    def _call_weights_dict(self, dic: Dict):
        if _is_callable(dic):
            if not _is_parametric(dic):
                _dic = copy.deepcopy(dic)
                for k, v in _dic.items():
                    if isinstance(_dic["cosine"][0], tuple):
                        val = _dic["cosine"][0][0]
                        if _is_callable(_dic["cosine"][0][0]):
                            val = _dic["cosine"][0][0]()
                        dur = _dic["cosine"][0][1]
                        if _is_callable(_dic["cosine"][0][1]):
                            dur = _dic["cosine"][0][1]()
                        _dic["cosine"][0] = (val, dur)

                    if isinstance(_dic["sine"][0], tuple):
                        val = _dic["sine"][0][0]
                        if _is_callable(_dic["sine"][0][0]):
                            val = _dic["sine"][0][0]()
                        dur = _dic["sine"][0][1]
                        if _is_callable(_dic["sine"][0][1]):
                            dur = _dic["sine"][0][1]()
                        _dic["sine"][0] = (val, dur)

                    elif _is_callable(_dic["cosine"]):
                        _dic["cosine"] = _dic["cosine"]()

                        if _is_callable(_dic["sine"]):
                            _dic["sine"] = _dic["sine"]()
                return _dic
        return dic

    def add_waveforms(self, wfs: List[Waveform]):
        """Add a list of waveforms to this configuration

        :param wfs: a list of Waveform objects
        :type wfs: List[Waveform]
        """
        for wf in wfs:
            self.add_waveform(wf)

    def add_pulse(self, pulse: Pulse):
        """Add a pulse to this configuration

        :param pulse: a Pulse object
        :type pulse: Pulse
        """
        _dict = copy.deepcopy(pulse.dict)
        self.add_waveforms(pulse.wfs)
        if pulse.digital_marker is not None:
            _dict["digital_marker"] = pulse.digital_marker.name
            self.config["digital_waveforms"][pulse.digital_marker.name] = self._call_dict_parameters(
                pulse.digital_marker.dict
            )
        if isinstance(pulse, MeasurePulse):
            _dict["integration_weights"] = dict()
            for w in pulse.integration_weights:
                _dict["integration_weights"][w.name] = w.weights.name
                self.config["integration_weights"][w.weights.name] = self._call_weights_dict(w.weights.dict)
        self.config["pulses"][pulse.name] = self._call_dict_parameters(_dict)

    def add_pulses(self, pulses: List[Pulse]):
        """Add a list of Pulses to this configuration

        :param pulses: a list of Pulse objects
        :type pulses: List[Pulse]
        """
        for pulse in pulses:
            self.add_pulse(pulse)

    def add_element(self, elm: Element):
        """Add a quantum element to this configuration

        :param elm: an object of type Element
        :type elm: Element
        """
        # check if input/output ports belong to the controller
        self.config["elements"][elm.name] = self._call_dict_parameters(elm.dict)
        if elm.type == "mixInputs":
            if elm.mixer is not None:
                self.add_mixer(elm.mixer)
            if elm.oscillator is not None:
                self.add_oscillator(elm.oscillator)
        self.add_pulses(elm.pulses)

    def update_pulse(self, elm_name: str, pulse: Pulse):
        """Change the pulse associated with an element to the pulse given

        :param elm_name: name of a quantum element in the configuration
        :type elm_name: str
        :param pulse: a Pulse object
        :type pulse: Pulse
        """
        assert elm_name in self.config["elements"].keys()
        self.config["elements"][elm_name]["operations"][pulse.name] = pulse.name
        self._add_pulse_to_config(pulse)

    def update_integration_weights(self, pulse_name: str, iw: IntegrationWeights):
        """Updates the integration_weights associated with a measurement pulse

        :param pulse_name: name of the measurement pulse
        :type pulse_name: str
        :param iw: IntegrationWeights object
        :type iw: IntegrationWeights
        """
        assert pulse_name in self.config["pulses"].keys()
        assert self.config["pulses"][pulse_name]["operation"] == "measure"
        assert len(self.config["pulses"][pulse_name]["waveforms"].keys()) == 2
        weight_name = pulse_name + "_weight"
        self.config["integration_weights"][weight_name] = self._call_dict_parameters(iw.dict)

    def update_constant_waveform(self, wf_name: str, amp: float):
        """Update the amplitude of an existing constant waveform

        :param wf_name: name of the existing waveform
        :type wf_name: str
        :param amp: amplitude to set in the range (-0.5 to 0.5)
        :type amp: float
        """
        assert wf_name in self.config["waveforms"].keys()
        assert self.config["waveforms"][wf_name]["type"] == "constant"
        self.config["waveforms"][wf_name]["sample"] = amp

    def update_arbitrary_waveform(self, wf_name: str, samples: List[float]):
        """Update the samples of an existing arbitrary waveform

        :param wf_name: name of the existing waveform
        :type wf_name: str
        :param samples: a vector of samples
        :type samples: List[float]
        """
        assert wf_name in self.config["waveforms"].keys()
        assert self.config["waveforms"][wf_name]["type"] == "arbitrary"
        self.config["waveforms"][wf_name]["samples"] = samples

    def reset(self, controllers=None):
        if controllers is None:
            controllers = []
        self.__init__(controllers)

    def update_intermediate_frequency(self, elm_name: str, freq: float, update_mixer: bool = True):
        self.config["elements"][elm_name]["intermediate_frequency"] = freq
        if update_mixer:
            m_name = self.config["elements"][elm_name]["mixer"]
            self.config["mixers"][m_name]["intermediate_frequency"] = freq

    def get_waveforms_from_op(self, elm_name, op_name):
        """Get waveforms associated with an operation

        :param elm_name: name of the element
        :type elm_name: [type]
        :param op_name: name of the operation
        :type op_name: [type]

        """
        assert elm_name in self.config["elements"].keys()
        assert op_name in self.config["elements"]["operations"].keys()
        waveforms = self.config["pulses"][op_name]["waveforms"]
        if waveforms.keys() == ["single"]:
            wf = waveforms["single"]
            _dict = self.config["waveforms"][wf]
            return [self.waveform_from_dict(wf, _dict)]
        else:
            wfI, wfQ = waveforms["I"], waveforms["Q"]
            _dictI = self.config["waveforms"][wfI]
            _dictQ = self.config["waveforms"][wfQ]
            return [
                self.waveform_from_dict(wfI, _dictI),
                self.waveform_from_dict(wfQ, _dictQ),
            ]

    def waveform_from_dict(self, wf, _dict):
        if _dict["type"] == "constant":
            return ConstantWaveform(wf, _dict["sample"])
        else:
            return ArbitraryWaveform(wf, _dict["samples"])

    def add_controller(self, cont: Controller):
        """Add a controller to this configuration

        :param cont: A Controller object
        :type cont: Controller
        """
        self.config["controllers"][cont.name] = cont.dict

    def add_integration_weights(self, weight: IntegrationWeights):
        """Add integration weights to this configuration

        :param weight: an IntegrationWeights object
        :type weight: IntegrationWeights
        """
        self.config["integration_weights"][weight.name] = self._call_dict_parameters(weight.dict)

    def add_mixer(self, mixer: Mixer):
        """Add a mixer to this configuration

        :param mixer: A Mixer object
        :type mixer: Mixer
        """
        self.config["mixers"][mixer.name] = [self._call_dict_parameters(data) for data in mixer.dict]

    def add_oscillator(self, oscillator: Oscillator):
        """Add an oscillator to this configuration

        :param oscillator: an Oscillator object
        :type oscillator: Oscillator
        """
        self.config["oscillators"][oscillator.name] = self._call_dict_parameters(oscillator.dict)



import base64
import datetime
import io
from typing import List, Optional
import matplotlib.pyplot as plt
import webbrowser

import importlib.resources as pkg_resources

from qualang_tools.wirer.instruments.instrument_channels import InstrumentChannels
from qualang_tools.wirer.visualizer.instrument_figure_manager import InstrumentFigureManager
from qualang_tools.wirer import visualizer


class WebInstrumentFigureManager(InstrumentFigureManager):
    """Extended figure manager that can export figures to web format"""

    def __init__(self):
        super().__init__()
        self.figure_objects = {}  # Store actual Figure objects

    def get_ax(self, con: int, slot: int, instrument_id: str):
        # Call parent method to get axes
        ax = super().get_ax(con, slot, instrument_id)

        # Store the figure object for later web export
        instrument_id_mapped = self.instrument_id_mapping.get(instrument_id, instrument_id)
        key = self._key(instrument_id_mapped, con)

        # Find and store the figure
        if isinstance(ax, dict):
            # For OPX1000 with multiple axes
            fig = list(ax.values())[0].figure
        else:
            # For single axis instruments
            fig = ax.figure

        self.figure_objects[key] = fig
        return ax

    def _key(self, instrument_id: str, con: int):
        return f"{instrument_id}_{con}"

    @property
    def instrument_id_mapping(self):
        from qualang_tools.wirer.visualizer.layout import instrument_id_mapping

        return instrument_id_mapping

    def export_figures_to_html(self, output_path: str = "instrument_visualization.html"):
        """Export all figures to a single HTML page"""

        # Convert figures to base64 encoded images
        figure_data = []

        for key, fig in self.figure_objects.items():
            # Remove suptitle to reduce image padding
            if fig._suptitle:
                fig._suptitle.remove()

            # Save figure to bytes buffer
            img_buffer = io.BytesIO()
            fig.patch.set_facecolor("#ffffff")  # Transparent figure background
            fig.savefig(img_buffer, format="png", dpi=150, bbox_inches="tight")
            img_buffer.seek(0)

            # Encode to base64
            img_base64 = base64.b64encode(img_buffer.read()).decode("utf-8")
            img_buffer.close()

            figure_data.append(
                {"key": key, "title": key, "data": img_base64}  # Use key as title since suptitle is removed
            )

        # Move OPX+ and OPX1000 figures to the end
        def priority(fig):
            title = fig["title"].lower()
            if "opx" in title:
                return 1  # Lower priority: appear later
            return 0  # Higher priority: appear first

        figure_data.sort(key=priority)

        # Generate HTML content
        html_content = self._generate_html(figure_data)

        # Write to file
        with open(output_path, "w", encoding="utf-8") as f:
            f.write(html_content)

        print(f"Visualization exported to: {output_path}")
        return output_path

    def _generate_html(self, figure_data: List[dict]) -> str:
        """Generate the HTML content with embedded figures"""

        with pkg_resources.files(visualizer).joinpath("web_template.html").open("r", encoding="utf-8") as f:
            html_template = f.read()

        # Generate figure HTML
        figures_html = '<div id="figs">'

        for i, fig_data in enumerate(figure_data):
            figures_html += f"""
            <div class="figure-container">
                <!-- Remove the caption and use the title as tooltip -->
                <p class="figure-caption">{fig_data['title'].replace('_', ' #')}</p>
                <img src="data:image/png;base64,{fig_data['data']}"
                     alt="{fig_data['title']}"
                     title="{fig_data['title'].replace('_', ' #')}"
                     class="figure-image"
                     style="transition: none;">
            </div>"""  # noqa: E272

        figures_html += "</div>"

        return html_template.format(figures_html=figures_html)


def visualize_web(
    qubit_dict,
    available_channels: Optional[InstrumentChannels] = None,
    output_path: str = None,
    show_matplotlib: bool = False,
):
    """
    Create web-based visualization of instrument wiring

    Args:
        qubit_dict: Dictionary of qubit configurations
        available_channels: Available instrument channels (optional)
        output_path: Path for output HTML file
        show_matplotlib: Whether to also show matplotlib windows

    Returns:
        str: Path to generated HTML file
    """
    # Import functions from your existing visualizer
    from qualang_tools.wirer.visualizer.visualizer import (
        invert_qubit_dict,
        make_annotations,
        merge_annotations_on_same_channel,
        make_unused_channel_annotations,
        draw_annotations,
    )

    # Process data same as original visualizer
    inverted_dict = invert_qubit_dict(qubit_dict)
    annotations = make_annotations(inverted_dict)
    annotations = merge_annotations_on_same_channel(annotations)

    # Use web-enabled figure manager
    manager = WebInstrumentFigureManager()

    # Draw unused channels if provided
    if available_channels is not None:
        available_channel_annotations = make_unused_channel_annotations(available_channels)
        draw_annotations(manager, available_channel_annotations)

    # Draw main annotations
    draw_annotations(manager, annotations)

    if output_path is None:
        output_path = f'qm-instrument_config_{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.html'

    # Export to HTML
    html_path = manager.export_figures_to_html(output_path)

    # Optionally show matplotlib windows too
    if show_matplotlib:
        plt.show()
    else:
        # Close matplotlib figures to save memory
        plt.close("all")

    return html_path


def visualize(qubit_dict, available_channels=None, use_matplotlib=None):
    """
    Args:
        qubit_dict: Dictionary of qubit configurations
        available_channels: Available instrument channels (optional)
        use_matplotlib: Whether to show matplotlib windows (new parameter) instead
    """
    if use_matplotlib:
        from qualang_tools.wirer.visualizer.visualizer import (
            invert_qubit_dict,
            make_annotations,
            merge_annotations_on_same_channel,
            make_unused_channel_annotations,
            draw_annotations,
        )

        # Replicate original visualize logic
        inverted_dict = invert_qubit_dict(qubit_dict)
        annotations = make_annotations(inverted_dict)
        annotations = merge_annotations_on_same_channel(annotations)

        manager = InstrumentFigureManager()

        if available_channels is not None:
            available_channel_annotations = make_unused_channel_annotations(available_channels)
            draw_annotations(manager, available_channel_annotations)

        draw_annotations(manager, annotations)
        plt.show()
    else:
        webbrowser.open(visualize_web(qubit_dict, available_channels, show_matplotlib=False))



from abc import ABC
from pathlib import Path


__all__ = ["DataProcessor"]


class DataProcessor(ABC):
    # Default separator for filename keys in the processed data
    nested_separator: str = "."

    def process(self, data):
        return data

    def post_process(self, data_folder: Path):
        pass



from pathlib import Path
import numpy as np

from .helpers import copy_nested_dict, iterate_nested_dict, update_nested_dict
from .data_processor import DataProcessor


class NumpyArraySaver(DataProcessor):
    merge_arrays: bool = True
    merged_array_name: str = "arrays.npz"
    nested_separator: str = "."

    def __init__(self, merge_arrays=None, merged_array_name=None):
        if merge_arrays is not None:
            self.merge_arrays = merge_arrays
        if merged_array_name is not None:
            self.merged_array_name = merged_array_name

        self.data_arrays = {}

    def process(self, data):
        self.data_arrays = {}
        processed_data = copy_nested_dict(data)

        for keys, val in iterate_nested_dict(data):
            if not isinstance(val, np.ndarray):
                continue

            path = Path(self.nested_separator.join(keys))
            self.data_arrays[path] = val
            if self.merge_arrays:
                update_nested_dict(processed_data, keys, f"./{self.merged_array_name}#{path}")
            else:
                update_nested_dict(processed_data, keys, f"./{path}.npy")
        return processed_data

    def post_process(self, data_folder: Path):
        if self.merge_arrays:
            arrays = {str(path): arr for path, arr in self.data_arrays.items()}
            if arrays:
                np.savez(data_folder / self.merged_array_name, **arrays)
        else:
            for path, arr in self.data_arrays.items():
                np.save(data_folder / f"{path}.npy", arr)
